from typing import Optional

from pydantic import BaseModel, Field


class RequestArtifactDefinitionGenerate(BaseModel):
    """Sent to trigger the generation of artifacts for a given branch."""

    artifact_definition_id: str = Field(..., description="The unique ID of the Artifact Definition")
    artifact_definition_name: str = Field(..., description="The name of the Artifact Definition")
    branch: str = Field(..., description="The branch to target")
    limit: list[str] = Field(
        default_factory=list,
        description="List of targets to limit the scope of the generation, if populated only the included artifacts will be regenerated",
    )


class RequestArtifactGenerate(BaseModel):
    """Runs to generate an artifact"""

    artifact_name: str = Field(..., description="Name of the artifact")
    artifact_definition: str = Field(..., description="The ID of the artifact definition")
    commit: str = Field(..., description="The commit to target")
    content_type: str = Field(..., description="Content type of the artifact")
    transform_type: str = Field(..., description="The type of transform associated with this artifact")
    transform_location: str = Field(..., description="The transforms location within the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    repository_kind: str = Field(..., description="The kind of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    target_id: str = Field(..., description="The ID of the target object for this artifact")
    target_name: str = Field(..., description="Name of the artifact target")
    artifact_id: Optional[str] = Field(default=None, description="The id of the artifact if it previously existed")
    query: str = Field(..., description="The name of the query to use when collecting data")
    timeout: int = Field(..., description="Timeout for requests used to generate this artifact")
    variables: dict = Field(..., description="Input variables when generating the artifact")


class GitRepositoryAdd(BaseModel):
    """Clone and sync an external repository after creation."""

    location: str = Field(..., description="The external URL of the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    created_by: Optional[str] = Field(default=None, description="The user ID of the user that created the repository")
    default_branch_name: Optional[str] = Field(None, description="Default branch for this repository")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")
    infrahub_branch_id: str = Field(..., description="Id of the Infrahub branch on which to sync the remote repository")
    internal_status: str = Field(..., description="Administrative status of the repository")


class GitRepositoryAddReadOnly(BaseModel):
    """Clone and sync an external repository after creation."""

    location: str = Field(..., description="The external URL of the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    ref: str = Field(..., description="Ref to track on the external repository")
    created_by: Optional[str] = Field(default=None, description="The user ID of the user that created the repository")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")
    infrahub_branch_id: str = Field(..., description="Id of the Infrahub branch on which to sync the remote repository")
    internal_status: str = Field(..., description="Internal status of the repository")


class GitRepositoryPullReadOnly(BaseModel):
    """Update a read-only repository to the latest commit for its ref"""

    location: str = Field(..., description="The external URL of the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    ref: Optional[str] = Field(None, description="Ref to track on the external repository")
    commit: Optional[str] = Field(None, description="Specific commit to pull")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")
    infrahub_branch_id: str = Field(..., description="Infrahub branch on which to sync the remote repository")


class GitRepositoryMerge(BaseModel):
    """Merge one branch into another."""

    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    internal_status: str = Field(..., description="Administrative status of the repository")
    source_branch: str = Field(..., description="The source branch")
    destination_branch: str = Field(..., description="The destination branch")
    destination_branch_id: str = Field(..., description="The ID of the destination branch")
    default_branch: str = Field(..., description="The default branch in Git")


class GitRepositoryImportObjects(BaseModel):
    """Re run import job against an existing commit."""

    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    repository_kind: str = Field(..., description="The type of repository")
    commit: str = Field(..., description="Specific commit to pull")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")


class GitDiffNamesOnly(BaseModel):
    """Request a list of modified files between two commits."""

    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    repository_kind: str = Field(..., description="The kind of the repository")
    first_commit: str = Field(..., description="The first commit")
    second_commit: Optional[str] = Field(None, description="The second commit")


class GitDiffNamesOnlyResponse(BaseModel):
    files_added: list[str] = Field(..., description="Files added")
    files_changed: list[str] = Field(..., description="Files changed")
    files_removed: list[str] = Field(..., description="Files removed")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field

from ..yaml import InfrahubFile, InfrahubFileKind

if TYPE_CHECKING:
    from ..client import InfrahubClient
    from ..schema import MainSchemaTypesAPI


class InfrahubObjectFileData(BaseModel):
    kind: str
    data: list[dict[str, Any]] = Field(default_factory=list)

    @classmethod
    def enrich_node(cls, data: dict, context: dict) -> dict:  # noqa: ARG003
        return data

    @classmethod
    async def create_node(
        cls,
        client: InfrahubClient,
        schema: MainSchemaTypesAPI,
        data: dict,
        context: dict | None = None,
        branch: str | None = None,
        default_schema_kind: str | None = None,
    ) -> None:
        # First validate of all mandatory fields are present
        for element in schema.mandatory_attribute_names + schema.mandatory_relationship_names:
            if element not in data.keys():
                raise ValueError(f"{element} is mandatory")

        clean_data: dict[str, Any] = {}

        remaining_rels = []
        for key, value in data.items():
            if key in schema.attribute_names:
                clean_data[key] = value

            if key in schema.relationship_names:
                rel_schema = schema.get_relationship(name=key)

                if isinstance(value, dict) and "data" not in value:
                    raise ValueError(f"Relationship {key} must be a dict with 'data'")

                # This is a simple implementation for now, need to revisit once we have the integration tests
                if isinstance(value, (list)):
                    clean_data[key] = value
                elif rel_schema.cardinality == "one" and isinstance(value, str):
                    clean_data[key] = [value]
                else:
                    remaining_rels.append(key)

        if context:
            clean_context = {
                ckey: cvalue
                for ckey, cvalue in context.items()
                if ckey in schema.relationship_names + schema.attribute_names
            }
            clean_data.update(clean_context)

        clean_data = cls.enrich_node(data=clean_data, context=context or {})

        node = await client.create(kind=schema.kind, branch=branch, data=clean_data)
        await node.save(allow_upsert=True)
        display_label = node.get_human_friendly_id_as_string() or f"{node.get_kind()} : {node.id}"
        client.log.info(f"Node: {display_label}")

        for rel in remaining_rels:
            # identify what is the name of the relationship on the other side
            if not isinstance(data[rel], dict) and "data" in data[rel]:
                raise ValueError(f"relationship {rel} must be a dict with 'data'")

            rel_schema = schema.get_relationship(name=rel)
            peer_kind = data[rel].get("kind", default_schema_kind) or rel_schema.peer
            peer_schema = await client.schema.get(kind=peer_kind, branch=branch)

            if rel_schema.identifier is None:
                raise ValueError("identifier must be defined")

            peer_rel = peer_schema.get_matching_relationship(id=rel_schema.identifier, direction=rel_schema.direction)

            rel_data = data[rel]["data"]
            context = {}
            if peer_rel:
                context[peer_rel.name] = node.id

            if rel_schema.cardinality == "one" and isinstance(rel_data, dict):
                await cls.create_node(
                    client=client,
                    schema=peer_schema,
                    data=rel_data,
                    context=context,
                    branch=branch,
                    default_schema_kind=default_schema_kind,
                )

            elif rel_schema.cardinality == "many" and isinstance(rel_data, list):
                for idx, peer_data in enumerate(rel_data):
                    context["list_index"] = idx
                    await cls.create_node(
                        client=client,
                        schema=peer_schema,
                        data=peer_data,
                        context=context,
                        branch=branch,
                        default_schema_kind=default_schema_kind,
                    )
            else:
                raise ValueError(
                    f"Relationship {rel_schema.name} doesn't have the right format {rel_schema.cardinality} / {type(rel_data)}"
                )


class ObjectFile(InfrahubFile):
    _spec: InfrahubObjectFileData | None = None

    @property
    def spec(self) -> InfrahubObjectFileData:
        if not self._spec:
            self._spec = InfrahubObjectFileData(**self.data.spec)
        return self._spec

    def validate_content(self) -> None:
        super().validate_content()
        if self.kind != InfrahubFileKind.OBJECT:
            raise ValueError("File is not an Infrahub Object file")
        self._spec = InfrahubObjectFileData(**self.data.spec)



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class FinalizeValidatorExecution(InfrahubMessage):
    """Update the status of a validator after all checks have been completed."""

    validator_id: str = Field(..., description="The id of the validator associated with this check")
    validator_execution_id: str = Field(..., description="The id of current execution of the associated validator")
    start_time: str = Field(..., description="Start time when the message was first created")
    validator_type: str = Field(..., description="The type of validator to complete")



from prefect import flow

from infrahub.core import registry
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.models import RequestDiffUpdate
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.dependencies.registry import get_component_registry
from infrahub.log import get_logger
from infrahub.services import services
from infrahub.workflows.catalogue import DIFF_REFRESH
from infrahub.workflows.utils import add_tags

log = get_logger()


@flow(name="diff-update", flow_run_name="Update diff for branch {model.branch_name}")
async def update_diff(model: RequestDiffUpdate) -> None:
    service = services.service
    await add_tags(branches=[model.branch_name])

    async with service.database.start_session() as db:
        component_registry = get_component_registry()
        base_branch = await registry.get_branch(db=db, branch=registry.default_branch)
        diff_branch = await registry.get_branch(db=db, branch=model.branch_name)

        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=diff_branch)

        await diff_coordinator.run_update(
            base_branch=base_branch,
            diff_branch=diff_branch,
            from_time=model.from_time,
            to_time=model.to_time,
            name=model.name,
        )


@flow(name="diff-refresh", flow_run_name="Recreate diff for branch {branch_name}")
async def refresh_diff(branch_name: str, diff_id: str) -> None:
    service = services.service
    await add_tags(branches=[branch_name])

    async with service.database.start_session() as db:
        component_registry = get_component_registry()
        base_branch = await registry.get_branch(db=db, branch=registry.default_branch)
        diff_branch = await registry.get_branch(db=db, branch=branch_name)

        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=diff_branch)
        await diff_coordinator.recalculate(base_branch=base_branch, diff_branch=diff_branch, diff_id=diff_id)


@flow(name="diff-refresh-all", flow_run_name="Recreate all diffs for branch {branch_name}")
async def refresh_diff_all(branch_name: str) -> None:
    service = services.service
    await add_tags(branches=[branch_name])

    async with service.database.start_session() as db:
        component_registry = get_component_registry()
        default_branch = registry.get_branch_from_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=default_branch)
        diff_roots_to_refresh = await diff_repository.get_roots_metadata(diff_branch_names=[branch_name])

        for diff_root in diff_roots_to_refresh:
            if diff_root.base_branch_name != diff_root.diff_branch_name:
                await service.workflow.submit_workflow(
                    workflow=DIFF_REFRESH,
                    parameters={"branch_name": diff_root.diff_branch_name, "diff_id": diff_root.uuid},
                )



from prefect import flow, task
from prefect.blocks.redis import RedisStorageContainer
from prefect.cache_policies import NONE
from prefect.client.orchestration import PrefectClient, get_client
from prefect.client.schemas.actions import WorkPoolCreate
from prefect.exceptions import ObjectAlreadyExists
from prefect.logging import get_run_logger

from infrahub import config

from .catalogue import automation_setup_workflows, worker_pools, workflows
from .models import TASK_RESULT_STORAGE_NAME


@task(name="task-manager-setup-worker-pools", task_run_name="Setup Worker pools", cache_policy=NONE)
async def setup_worker_pools(client: PrefectClient) -> None:
    log = get_run_logger()
    for worker in worker_pools:
        wp = WorkPoolCreate(
            name=worker.name,
            type=worker.worker_type or config.SETTINGS.workflow.default_worker_type,
            description=worker.description,
        )
        try:
            await client.create_work_pool(work_pool=wp, overwrite=True)
            log.info(f"Work pool {worker.name} created successfully ... ")
        except ObjectAlreadyExists:
            log.warning(f"Work pool {worker.name} already present ")


@task(name="task-manager-setup-deployments", task_run_name="Setup Deployments", cache_policy=NONE)
async def setup_deployments(client: PrefectClient) -> None:
    log = get_run_logger()
    for workflow in workflows:
        # For now the workpool is hardcoded but
        # later we need to make it dynamic to have a different worker based on the type of the workflow
        work_pool = worker_pools[0]
        await workflow.save(client=client, work_pool=work_pool)
        log.info(f"Flow {workflow.name}, created successfully ... ")

    for automation_setup_workflow in automation_setup_workflows:
        automation_setup = automation_setup_workflow.get_function()
        await automation_setup()


@task(name="task-manager-setup-blocks", task_run_name="Setup Blocks", cache_policy=NONE)
async def setup_blocks() -> None:
    log = get_run_logger()

    try:
        await RedisStorageContainer.register_type_and_schema()
    except ObjectAlreadyExists:
        log.warning(f"Redis Storage {TASK_RESULT_STORAGE_NAME} already registered ")

    redis_block = RedisStorageContainer.from_host(
        host=config.SETTINGS.cache.address,
        port=config.SETTINGS.cache.service_port,
        db=config.SETTINGS.cache.database,
        username=config.SETTINGS.cache.username or None,
        password=config.SETTINGS.cache.password or None,
    )
    try:
        await redis_block.save(name=TASK_RESULT_STORAGE_NAME, overwrite=True)
    except ObjectAlreadyExists:
        log.warning(f"Redis Storage {TASK_RESULT_STORAGE_NAME} already present ")


@flow(name="task-manager-setup", flow_run_name="Setup Task Manager")
async def setup_task_manager() -> None:
    async with get_client(sync_client=False) as client:
        await setup_blocks()
        await setup_worker_pools(client=client)
        await setup_deployments(client=client)



import enum
import re
from typing import Any

ENUM_NAME_REGEX = re.compile("[_a-zA-Z0-9]+")


def generate_python_enum(name: str, options: list[Any]) -> type[enum.Enum]:
    main_attrs = {}
    for option in options:
        if isinstance(option, int):
            enum_name = str(option)
        else:
            enum_name = "_".join(re.findall(ENUM_NAME_REGEX, option)).upper()

        main_attrs[enum_name] = option
    return enum.Enum(name, main_attrs)  # type: ignore[return-value]



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class EventWorkerNewPrimaryAPI(InfrahubMessage):
    """Sent on startup or when a new primary API worker is elected."""

    worker_id: str = Field(..., description="The worker ID that got elected")



from infrahub.core.validators.attribute.optional import AttributeOptionalChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaAttributeOptionalConstraintDependency(DependencyBuilder[AttributeOptionalChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AttributeOptionalChecker:
        return AttributeOptionalChecker(db=context.db, branch=context.branch)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Sequence

from infrahub.core import registry
from infrahub.core.diff.payload_builder import get_display_labels_per_kind
from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.schema import GenericSchema, NodeSchema, SchemaRoot, internal_schema
from infrahub.core.schema.manager import SchemaManager
from infrahub.core.validators.uniqueness.checker import UniquenessChecker
from infrahub.dependencies.registry import build_component_registry, get_component_registry
from infrahub.log import get_logger

from ..shared import InternalSchemaMigration, SchemaMigration

if TYPE_CHECKING:
    from infrahub.core.validators.uniqueness.model import NonUniqueNode
    from infrahub.database import InfrahubDatabase

log = get_logger()


class Migration018(InternalSchemaMigration):
    name: str = "018_validate_nulls_in_uniqueness_constraints"
    minimum_version: int = 17
    migrations: Sequence[SchemaMigration] = []

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()

        return result

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        """
        Validate any schema that include optional attributes in the uniqueness constraints

        An update to uniqueness constraint validation now handles NULL values as unique instead of ignoring them
        """
        default_branch = registry.get_branch_from_registry()
        build_component_registry()
        component_registry = get_component_registry()
        uniqueness_checker = await component_registry.get_component(UniquenessChecker, db=db, branch=default_branch)
        non_unique_nodes_by_kind: dict[str, list[NonUniqueNode]] = defaultdict(list)

        manager = SchemaManager()
        registry.schema = manager
        internal_schema_root = SchemaRoot(**internal_schema)
        manager.register_schema(schema=internal_schema_root)
        schema_branch = await manager.load_schema_from_db(db=db, branch=default_branch)
        manager.set_schema_branch(name=default_branch.name, schema=schema_branch)

        for schema_kind in schema_branch.node_names + schema_branch.generic_names:
            schema = schema_branch.get(name=schema_kind, duplicate=False)
            if not isinstance(schema, (NodeSchema, GenericSchema)):
                continue

            schema_constraint_path_groups = schema.get_unique_constraint_schema_attribute_paths(
                schema_branch=schema_branch
            )
            includes_optional_attr: bool = False

            for constraint_group in schema_constraint_path_groups:
                for schema_attribute_path in constraint_group:
                    if (
                        schema_attribute_path.attribute_schema
                        and schema_attribute_path.attribute_schema.optional is True
                    ):
                        includes_optional_attr = True
                        break

            if not includes_optional_attr:
                continue

            non_unique_nodes = await uniqueness_checker.check_one_schema(schema=schema)
            if non_unique_nodes:
                non_unique_nodes_by_kind[schema_kind] = non_unique_nodes

        if not non_unique_nodes_by_kind:
            return MigrationResult()

        error_strings = []
        for schema_kind, non_unique_nodes in non_unique_nodes_by_kind.items():
            display_label_map = await get_display_labels_per_kind(
                db=db, kind=schema_kind, branch_name=default_branch.name, ids=[nun.node_id for nun in non_unique_nodes]
            )
            for non_unique_node in non_unique_nodes:
                display_label = display_label_map.get(non_unique_node.node_id)
                error_str = f"{display_label or ''}({non_unique_node.node_schema.kind} / {non_unique_node.node_id})"
                error_str += " violates uniqueness constraints for the following attributes: "
                attr_values = [
                    f"{attr.attribute_name}={attr.attribute_value}" for attr in non_unique_node.non_unique_attributes
                ]
                error_str += ", ".join(attr_values)
                error_strings.append(error_str)
        if error_strings:
            error_str = "For the following nodes, you must update the uniqueness_constraints on the schema of the node"
            error_str += " to remove the attribute(s) with NULL values or update the data on the nodes to be unique"
            error_str += " now that NULL values are considered during uniqueness validation"
            return MigrationResult(errors=[error_str] + error_strings)
        return MigrationResult()



from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from .deserializer import DiffDeserializerDependency


class DiffRepositoryDependency(DependencyBuilder[DiffRepository]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffRepository:
        return DiffRepository(db=context.db, deserializer=DiffDeserializerDependency.build(context=context))



import logging

import typer
from infrahub_sdk.async_typer import AsyncTyper
from prefect.client.orchestration import get_client

from infrahub import config
from infrahub.services.adapters.workflow.worker import WorkflowWorkerExecution
from infrahub.tasks.dummy import DUMMY_FLOW, DummyInput
from infrahub.workflows.initialization import setup_task_manager
from infrahub.workflows.models import WorkerPoolDefinition

app = AsyncTyper()

# pylint: disable=unused-argument


@app.command()
async def init(
    ctx: typer.Context,
    debug: bool = typer.Option(False, help="Enable advanced logging and troubleshooting"),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Initialize the task manager"""
    logging.getLogger("prefect").setLevel(logging.ERROR)

    config.load_and_exit(config_file_name=config_file)

    await setup_task_manager()


@app.command()
async def execute(
    ctx: typer.Context,
    debug: bool = typer.Option(False, help="Enable advanced logging and troubleshooting"),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Check the current format of the internal graph and apply the necessary migrations"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    config.load_and_exit(config_file_name=config_file)

    async with get_client(sync_client=False) as client:
        worker = WorkflowWorkerExecution()
        await DUMMY_FLOW.save(
            client=client, work_pool=WorkerPoolDefinition(name="infrahub-worker", worker_type="infrahubasync")
        )

        result = await worker.execute_workflow(
            workflow=DUMMY_FLOW, parameters={"data": DummyInput(firstname="John", lastname="Doe")}
        )  # type: ignore[var-annotated]
        print(result)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import ID, Field, List, String
from infrahub_sdk.utils import extract_fields_first_node

from infrahub.graphql.types import BranchType

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo


async def branch_resolver(
    root: dict,  # pylint: disable=unused-argument
    info: GraphQLResolveInfo,
    **kwargs: Any,
) -> list[dict[str, Any]]:
    fields = await extract_fields_first_node(info)
    return await BranchType.get_list(context=info.context, fields=fields, **kwargs)


BranchQueryList = Field(
    List(BranchType),
    ids=List(ID),
    name=String(),
    resolver=branch_resolver,
)



from __future__ import annotations

import asyncio
import random
from dataclasses import dataclass
from typing import TYPE_CHECKING

from infrahub import config
from infrahub.components import ComponentType
from infrahub.tasks.keepalive import refresh_heartbeat
from infrahub.tasks.recurring import trigger_branch_refresh

if TYPE_CHECKING:
    from infrahub.services import InfrahubServices, ServiceFunction


@dataclass
class Schedule:
    name: str
    interval: int
    function: ServiceFunction
    start_delay: int = 30


class InfrahubScheduler:
    def __init__(self) -> None:
        self.service: InfrahubServices
        self.running: bool = False
        self.schedules: list[Schedule] = []

    async def initialize(self, service: InfrahubServices) -> None:
        self.service = service

        self.running = config.SETTINGS.miscellaneous.start_background_runner
        # Add some randomness to the interval to avoid having all workers pulling the latest update at the same time
        random_number = random.randint(0, 5)
        if self.service.component_type == ComponentType.API_SERVER:
            schedules = [
                Schedule(name="refresh_api_components", interval=10, function=refresh_heartbeat, start_delay=0),
                Schedule(
                    name="branch_refresh", interval=900, function=trigger_branch_refresh, start_delay=random_number
                ),
            ]
            self.schedules.extend(schedules)

        if self.service.component_type == ComponentType.GIT_AGENT:
            schedules = [
                Schedule(name="refresh_components", interval=10, function=refresh_heartbeat),
                Schedule(
                    name="branch_refresh", interval=900, function=trigger_branch_refresh, start_delay=random_number
                ),
            ]
            self.schedules.extend(schedules)

    async def start_schedule(self) -> None:
        for schedule in self.schedules:
            asyncio.create_task(
                run_schedule(schedule=schedule, service=self.service), name=f"scheduled_task_{schedule.name}"
            )

    async def shutdown(self) -> None:
        self.running = False


async def run_schedule(schedule: Schedule, service: InfrahubServices) -> None:
    """Execute the task provided in the schedule as per the defined interval

    Once the service is marked to be shutdown the scheduler will stop executing tasks.
    """
    for _ in range(schedule.start_delay):
        if not service.scheduler.running:
            return
        await asyncio.sleep(delay=1)

    service.log.info("Started recurring task", task=schedule.name)

    while service.scheduler.running:
        try:
            await schedule.function(service)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            service.log.error(str(exc))
        for _ in range(schedule.interval):
            if not service.scheduler.running:
                return
            await asyncio.sleep(delay=1)



from infrahub.core.diff.conflicts_enricher import ConflictsEnricher
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffConflictsEnricherDependency(DependencyBuilder[ConflictsEnricher]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> ConflictsEnricher:
        return ConflictsEnricher()



from __future__ import annotations

from typing import TYPE_CHECKING

from graphene import Boolean, InputField, InputObjectType, List, Mutation, String
from infrahub_sdk.utils import compare_lists

from infrahub.core.account import GlobalPermission, ObjectPermission
from infrahub.core.constants import InfrahubKind, PermissionAction, PermissionDecision, RelationshipCardinality
from infrahub.core.manager import NodeManager
from infrahub.core.query.relationship import (
    RelationshipGetPeerQuery,
    RelationshipPeerData,
)
from infrahub.core.relationship import Relationship
from infrahub.database import retry_db_transaction
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.permissions import get_global_permission_for_kind

from ..types import RelatedNodeInput

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.relationship import RelationshipManager

    from ..initialization import GraphqlContext


# pylint: disable=unused-argument,too-many-branches

RELATIONSHIP_PEERS_TO_IGNORE = [InfrahubKind.NODE]


class RelationshipNodesInput(InputObjectType):
    id = InputField(String(required=True), description="ID of the node at the source of the relationship")
    name = InputField(String(required=True), description="Name of the relationship to add or remove nodes")
    nodes = InputField(
        List(of_type=RelatedNodeInput), description="List of nodes to add or remove to the relationships"
    )


class RelationshipMixin:
    @classmethod
    async def mutate(
        cls,
        root: dict,
        info: GraphQLResolveInfo,
        data: RelationshipNodesInput,
    ):
        context: GraphqlContext = info.context
        input_id = str(data.id)

        if not (
            source := await NodeManager.get_one(
                db=context.db,
                id=input_id,
                branch=context.branch,
                include_owner=True,
                include_source=True,
            )
        ):
            raise NodeNotFoundError(context.branch, None, input_id)

        # Check if the name of the relationship provided exist for this node and is of cardinality Many
        if data.get("name") not in source._schema.relationship_names:
            raise ValidationError(
                {"name": f"'{data.get('name')}' is not a valid relationship for '{source.get_kind()}'"}
            )

        rel_schema = source._schema.get_relationship(name=data.get("name"))
        if rel_schema.cardinality != RelationshipCardinality.MANY:
            raise ValidationError({"name": f"'{data.get('name')}' must be a relationship of cardinality Many"})

        # Query the node in the database and validate that all of them exist and are if the correct kind
        node_ids: list[str] = [node_data["id"] for node_data in data.get("nodes") if "id" in node_data]
        nodes = await NodeManager.get_many(
            db=context.db, ids=node_ids, fields={"display_label": None}, branch=context.branch
        )

        if context.account_session:
            impacted_schemas = {node.get_schema() for node in [source] + list(nodes.values())}
            required_permissions: list[GlobalPermission | ObjectPermission] = []
            decision = (
                PermissionDecision.ALLOW_DEFAULT.value
                if context.branch.is_default
                else PermissionDecision.ALLOW_OTHER.value
            )

            for impacted_schema in impacted_schemas:
                global_action = get_global_permission_for_kind(schema=impacted_schema)

                if global_action:
                    required_permissions.append(GlobalPermission(action=global_action, decision=decision))
                else:
                    required_permissions.append(
                        ObjectPermission(
                            namespace=impacted_schema.namespace,
                            name=impacted_schema.name,
                            action=PermissionAction.UPDATE.value,
                            decision=decision,
                        )
                    )

            context.active_permissions.raise_for_permissions(permissions=required_permissions)

        _, _, in_list2 = compare_lists(list1=list(nodes.keys()), list2=node_ids)
        if in_list2:
            for node_id in in_list2:
                raise ValidationError(f"{node_id!r}: Unable to find the node in the database.")

        for node_id, node in nodes.items():
            if rel_schema.peer in RELATIONSHIP_PEERS_TO_IGNORE:
                continue
            if rel_schema.peer not in node.get_labels():
                raise ValidationError(f"{node_id!r} {node.get_kind()!r} is not a valid peer for '{rel_schema.peer}'")

            peer_relationships = [rel for rel in node._schema.relationships if rel.identifier == rel_schema.identifier]
            if (
                rel_schema.identifier
                and len(peer_relationships) == 1
                and peer_relationships[0].cardinality == RelationshipCardinality.ONE
            ):
                peer_relationship: RelationshipManager = getattr(node, peer_relationships[0].name)
                if peer := await peer_relationship.get_peer(db=context.db):
                    if peer.id != input_id:
                        raise ValidationError(
                            f"{node_id!r} {node.get_kind()!r} is already related to another peer on '{peer_relationships[0].name}'"
                        )

        # The nodes that are already present in the db
        query = await RelationshipGetPeerQuery.init(
            db=context.db,
            source=source,
            rel=Relationship(schema=rel_schema, branch=context.branch, node=source),
        )
        await query.execute(db=context.db)
        existing_peers: dict[str, RelationshipPeerData] = {peer.peer_id: peer for peer in query.get_peers()}

        async with context.db.start_transaction() as db:
            if cls.__name__ == "RelationshipAdd":
                for node_data in data.get("nodes"):
                    # Instantiate and resolve a relationship
                    # This will take care of allocating a node from a pool if needed
                    rel = Relationship(schema=rel_schema, branch=context.branch, node=source)
                    await rel.new(db=db, data=node_data)
                    await rel.resolve(db=db)
                    # Save it only if it does not exist
                    if rel.get_peer_id() not in existing_peers.keys():
                        await rel.save(db=db)

            elif cls.__name__ == "RelationshipRemove":
                for node_data in data.get("nodes"):
                    if node_data.get("id") in existing_peers.keys():
                        # TODO once https://github.com/opsmill/infrahub/issues/792 has been fixed
                        # we should use RelationshipDataDeleteQuery to delete the relationship
                        # it would be more query efficient
                        rel = Relationship(schema=rel_schema, branch=context.branch, node=source)
                        await rel.load(db=db, data=existing_peers[node_data.get("id")])
                        await rel.delete(db=db)

        return cls(ok=True)


class RelationshipAdd(RelationshipMixin, Mutation):
    class Arguments:
        data = RelationshipNodesInput(required=True)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="relationship_add")
    async def mutate(
        cls,
        root: dict,
        info: GraphQLResolveInfo,
        data: RelationshipNodesInput,
    ):
        return await super().mutate(root=root, info=info, data=data)


class RelationshipRemove(RelationshipMixin, Mutation):
    class Arguments:
        data = RelationshipNodesInput(required=True)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="relationship_remove")
    async def mutate(
        cls,
        root: dict,
        info: GraphQLResolveInfo,
        data: RelationshipNodesInput,
    ):
        return await super().mutate(root=root, info=info, data=data)



from prefect import flow

from infrahub.message_bus import messages
from infrahub.services import InfrahubServices
from infrahub.workflows.catalogue import WEBHOOK_CONFIGURE


@flow(name="event-worker-newprimary-api")
async def new_primary_api(message: messages.EventWorkerNewPrimaryAPI, service: InfrahubServices) -> None:
    service.log.info("api_worker promoted to primary", worker_id=message.worker_id)

    await service.workflow.submit_workflow(workflow=WEBHOOK_CONFIGURE)



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import InfrahubDatabase
    from .index import IndexManagerBase


class DatabaseManager(ABC):
    index: IndexManagerBase

    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from infrahub.core.query.resource_manager import NumberPoolGetReserved, NumberPoolGetUsed, NumberPoolSetReserved
from infrahub.exceptions import PoolExhaustedError

from .. import Node

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase


class CoreNumberPool(Node):
    async def get_resource(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        node: Node,
        identifier: Optional[str] = None,
    ) -> int:
        identifier = identifier or node.get_id()
        # Check if there is already a resource allocated with this identifier
        # if not, pull all existing prefixes and allocated the next available
        # TODO add support for branch, if the node is reserved with this id in another branch we should return an error
        query_get = await NumberPoolGetReserved.init(db=db, branch=branch, pool_id=self.id, identifier=identifier)
        await query_get.execute(db=db)
        reservation = query_get.get_reservation()
        if reservation is not None:
            return reservation

        # If we have not returned a value we need to find one if avaiable
        number = await self.get_next(db=db, branch=branch)

        query_set = await NumberPoolSetReserved.init(
            db=db, pool_id=self.get_id(), identifier=identifier, reserved=number
        )
        await query_set.execute(db=db)

        return number

    async def get_next(self, db: InfrahubDatabase, branch: Branch) -> int:
        query = await NumberPoolGetUsed.init(db=db, branch=branch, pool=self, branch_agnostic=True)
        await query.execute(db=db)
        taken = [result.get_as_optional_type("av.value", return_type=int) for result in query.results]
        next_number = find_next_free(
            start=self.start_range.value,  # type: ignore[attr-defined]
            end=self.end_range.value,  # type: ignore[attr-defined]
            taken=taken,
        )
        if next_number is None:
            raise PoolExhaustedError("There are no more addresses available in this pool.")

        return next_number


def find_next_free(start: int, end: int, taken: list[int | None]) -> Optional[int]:
    used_numbers = [number for number in taken if number is not None]
    used_set = set(used_numbers)

    for num in range(start, end + 1):
        if num not in used_set:
            return num

    return None



import uuid
from typing import Any

from infrahub.workflows.models import WorkflowDefinition, WorkflowInfo

from . import InfrahubWorkflow, Return


class WorkflowLocalExecution(InfrahubWorkflow):
    async def execute_workflow(
        self,
        workflow: WorkflowDefinition,
        expected_return: type[Return] | None = None,
        parameters: dict[str, Any] | None = None,
        tags: list[str] | None = None,
    ) -> Any:
        fn = workflow.get_function()
        return await fn(**parameters or {})

    async def submit_workflow(
        self,
        workflow: WorkflowDefinition,
        parameters: dict[str, Any] | None = None,
        tags: list[str] | None = None,
    ) -> WorkflowInfo:
        await self.execute_workflow(workflow=workflow, parameters=parameters)
        return WorkflowInfo(id=uuid.uuid4())



import importlib
from uuid import uuid4

from infrahub import config, lock
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import (
    DEFAULT_IP_NAMESPACE,
    GLOBAL_BRANCH_NAME,
    GlobalPermissions,
    InfrahubKind,
    PermissionAction,
    PermissionDecision,
)
from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.node.ipam import BuiltinIPPrefix
from infrahub.core.node.permissions import CoreGlobalPermission, CoreObjectPermission
from infrahub.core.node.resource_manager.ip_address_pool import CoreIPAddressPool
from infrahub.core.node.resource_manager.ip_prefix_pool import CoreIPPrefixPool
from infrahub.core.node.resource_manager.number_pool import CoreNumberPool
from infrahub.core.protocols import CoreAccount
from infrahub.core.root import Root
from infrahub.core.schema import SchemaRoot, core_models, internal_schema
from infrahub.core.schema.manager import SchemaManager
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import DatabaseError
from infrahub.graphql.manager import GraphQLSchemaManager
from infrahub.log import get_logger
from infrahub.menu.menu import default_menu
from infrahub.menu.utils import create_menu_children
from infrahub.permissions import PermissionBackend
from infrahub.storage import InfrahubObjectStorage

log = get_logger()


async def get_root_node(db: InfrahubDatabase, initialize: bool = False) -> Root:
    roots = await Root.get_list(db=db)
    if len(roots) == 0 and not initialize:
        raise DatabaseError(
            "The Database hasn't been initialized for Infrahub, please run 'infrahub db init' or 'infrahub server start' to initialize the database."
        )

    if len(roots) == 0:
        await first_time_initialization(db=db)
        roots = await Root.get_list(db=db)

    elif len(roots) > 1:
        raise DatabaseError("The Database is corrupted, more than 1 root node found.")

    return roots[0]


async def get_default_ipnamespace(db: InfrahubDatabase) -> Node | None:
    if not registry.schema._branches or not registry.schema.has(name=InfrahubKind.NAMESPACE):
        return None

    nodes = await registry.manager.query(db=db, schema=InfrahubKind.NAMESPACE, filters={"default__value": True})
    if len(nodes) == 0:
        return None

    if len(nodes) > 1:
        raise DatabaseError("More than 1 default namespace found.")

    return nodes[0]


def initialize_permission_backends() -> list[PermissionBackend]:
    permission_backends: list[PermissionBackend] = []
    for backend_module_path in config.SETTINGS.main.permission_backends:
        log.info("Loading permission backend", backend=backend_module_path)

        module, class_name = backend_module_path.rsplit(".", maxsplit=1)
        Backend = getattr(importlib.import_module(module), class_name)
        permission_backends.append(Backend())

    return permission_backends


async def initialize_registry(db: InfrahubDatabase, initialize: bool = False) -> None:
    # ---------------------------------------------------
    # Initialize the database and Load the Root node
    # ---------------------------------------------------
    root = await get_root_node(db=db, initialize=initialize)
    registry.id = str(root.get_uuid())
    registry.default_branch = root.default_branch

    # ---------------------------------------------------
    # Initialize the Storage Driver
    # ---------------------------------------------------
    registry.storage = await InfrahubObjectStorage.init(settings=config.SETTINGS.storage)

    # ---------------------------------------------------
    # Load existing branches into the registry
    # ---------------------------------------------------
    branches: list[Branch] = await Branch.get_list(db=db)
    for branch in branches:
        registry.branch[branch.name] = branch

    # ---------------------------------------------------
    # Load internal models into the registry
    # ---------------------------------------------------
    registry.node["Node"] = Node
    registry.node[InfrahubKind.IPPREFIX] = BuiltinIPPrefix
    registry.node[InfrahubKind.IPADDRESSPOOL] = CoreIPAddressPool
    registry.node[InfrahubKind.IPPREFIXPOOL] = CoreIPPrefixPool
    registry.node[InfrahubKind.NUMBERPOOL] = CoreNumberPool
    registry.node[InfrahubKind.GLOBALPERMISSION] = CoreGlobalPermission
    registry.node[InfrahubKind.OBJECTPERMISSION] = CoreObjectPermission

    # ---------------------------------------------------
    # Instantiate permission backends
    # ---------------------------------------------------
    registry.permission_backends = initialize_permission_backends()


async def initialization(db: InfrahubDatabase) -> None:
    if config.SETTINGS.database.db_type == config.DatabaseType.MEMGRAPH:
        session = await db.session()
        await session.run(query="SET DATABASE SETTING 'log.level' TO 'INFO'")
        await session.run(query="SET DATABASE SETTING 'log.to_stderr' TO 'true'")
        await session.run(query="STORAGE MODE IN_MEMORY_ANALYTICAL")

    # ---------------------------------------------------
    # Initialize the database and Load the Root node
    # ---------------------------------------------------
    async with lock.registry.initialization():
        log.debug("Checking Root Node")
        await initialize_registry(db=db, initialize=True)

        # Add Indexes to the database
        if db.manager.index.initialized:
            log.debug("Loading database indexes ..")
            await db.manager.index.add()
        else:
            log.warning("The database index manager hasn't been initialized.")

    # ---------------------------------------------------
    # Load all schema in the database into the registry
    #  ... Unless the schema has been initialized already
    # ---------------------------------------------------
    if not registry.schema_has_been_initialized():
        registry.schema = SchemaManager()
        schema = SchemaRoot(**internal_schema)
        registry.schema.register_schema(schema=schema)

        # Import the default branch
        default_branch: Branch = registry.get_branch_from_registry(branch=registry.default_branch)
        hash_in_db = default_branch.active_schema_hash.main
        schema_default_branch = await registry.schema.load_schema_from_db(db=db, branch=default_branch)
        registry.schema.set_schema_branch(name=default_branch.name, schema=schema_default_branch)

        if default_branch.update_schema_hash():
            log.warning(
                "New schema detected after pulling the schema from the db",
                hash_current=hash_in_db,
                hash_new=default_branch.active_schema_hash.main,
                branch=default_branch.name,
            )
            await default_branch.save(db=db)

        for branch in list(registry.branch.values()):
            if branch.name in [default_branch.name, GLOBAL_BRANCH_NAME]:
                continue

            hash_in_db = branch.active_schema_hash.main
            log.info("Importing schema", branch=branch.name)
            await registry.schema.load_schema(db=db, branch=branch)

            if branch.update_schema_hash():
                log.warning(
                    f"New schema detected after pulling the schema from the db :"
                    f" {hash_in_db!r} >> {branch.active_schema_hash.main!r}",
                    branch=branch.name,
                )
                await branch.save(db=db)

    default_branch = registry.get_branch_from_registry(branch=registry.default_branch)
    schema_branch = registry.schema.get_schema_branch(name=default_branch.name)
    gqlm = GraphQLSchemaManager.get_manager_for_branch(branch=default_branch, schema_branch=schema_branch)
    gqlm.get_graphql_schema(
        include_query=True,
        include_mutation=True,
        include_subscription=True,
        include_types=True,
    )

    # ---------------------------------------------------
    # Load Default Namespace
    # ---------------------------------------------------
    ip_namespace = await get_default_ipnamespace(db=db)
    if ip_namespace:
        registry.default_ipnamespace = ip_namespace.id


async def create_root_node(db: InfrahubDatabase) -> Root:
    root = Root(graph_version=GRAPH_VERSION, default_branch=config.SETTINGS.initial.default_branch)
    await root.save(db=db)
    log.info(f"Generated instance ID : {root.uuid} (v{GRAPH_VERSION})")

    registry.id = root.id
    registry.default_branch = root.default_branch

    return root


async def create_default_branch(db: InfrahubDatabase) -> Branch:
    branch = Branch(
        name=registry.default_branch,
        status="OPEN",
        description="Default Branch",
        hierarchy_level=1,
        is_default=True,
        sync_with_git=True,
    )
    await branch.save(db=db)
    registry.branch[branch.name] = branch

    log.info("Created default branch", branch=branch.name)

    return branch


async def create_global_branch(db: InfrahubDatabase) -> Branch:
    branch = Branch(
        name=GLOBAL_BRANCH_NAME,
        status="OPEN",
        description="Global Branch",
        hierarchy_level=1,
        is_global=True,
        sync_with_git=False,
    )
    await branch.save(db=db)
    registry.branch[branch.name] = branch

    log.info("Created global branch", branch=branch.name)

    return branch


async def create_branch(
    branch_name: str, db: InfrahubDatabase, description: str = "", isolated: bool = True, at: str | None = None
) -> Branch:
    """Create a new Branch, currently all the branches are based on Main

    Because all branches are based on main, the hierarchy_level of hardcoded to 2."""
    description = description or f"Branch {branch_name}"
    branch = Branch(
        name=branch_name,
        status="OPEN",
        hierarchy_level=2,
        description=description,
        is_default=False,
        sync_with_git=False,
        created_at=at,
        branched_from=at,
        is_isolated=isolated,
    )

    origin_schema = registry.schema.get_schema_branch(name=branch.origin_branch)
    new_schema = origin_schema.duplicate(name=branch.name)
    registry.schema.set_schema_branch(name=branch.name, schema=new_schema)

    branch.update_schema_hash()
    await branch.save(db=db)
    registry.branch[branch.name] = branch

    log.info("Created branch", branch=branch.name)

    return branch


async def create_account(
    db: InfrahubDatabase,
    name: str = "admin",
    password: str | None = None,
    token_value: str | None = None,
) -> CoreAccount:
    token_schema = db.schema.get_node_schema(name=InfrahubKind.ACCOUNTTOKEN)
    obj = await Node.init(db=db, schema=CoreAccount)
    await obj.new(db=db, name=name, account_type="User", password=password)
    await obj.save(db=db)
    log.info(f"Created Account: {name}", account_name=name)

    if token_value:
        token = await Node.init(db=db, schema=token_schema)
        await token.new(db=db, token=token_value, name="Created automatically", account=obj)
        await token.save(db=db)

    return obj


async def create_ipam_namespace(
    db: InfrahubDatabase,
    name: str = DEFAULT_IP_NAMESPACE,
    description: str = "Used to provide a default space of IP resources",
) -> Node:
    obj = await Node.init(db=db, schema=InfrahubKind.NAMESPACE)
    await obj.new(db=db, name=name, description=description, default=True)
    await obj.save(db=db)
    log.info(f"Created IPAM Namespace: {name}")

    return obj


async def create_default_menu(db: InfrahubDatabase) -> None:
    for item in default_menu:
        obj = await item.to_node(db=db)
        await obj.save(db=db)
        if item.children:
            await create_menu_children(db=db, parent=obj, children=item.children)


async def create_super_administrator_role(db: InfrahubDatabase) -> Node:
    permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
    await permission.new(
        db=db,
        action=GlobalPermissions.SUPER_ADMIN.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to do anything",
    )
    await permission.save(db=db)
    log.info(f"Created global permission: {GlobalPermissions.SUPER_ADMIN}")

    role_name = "Super Administrator"
    obj = await Node.init(db=db, schema=InfrahubKind.ACCOUNTROLE)
    await obj.new(db=db, name=role_name, permissions=[permission])
    await obj.save(db=db)
    log.info(f"Created account role: {role_name}")

    return obj


async def create_default_roles(db: InfrahubDatabase) -> Node:
    repo_permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
    await repo_permission.new(
        db=db,
        action=GlobalPermissions.MANAGE_REPOSITORIES.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to manage repositories",
    )
    await repo_permission.save(db=db)

    schema_permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
    await schema_permission.new(
        db=db,
        action=GlobalPermissions.MANAGE_SCHEMA.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to manage the schema",
    )
    await schema_permission.save(db=db)

    proposed_change_permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
    await proposed_change_permission.new(
        db=db,
        action=GlobalPermissions.MERGE_PROPOSED_CHANGE.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to merge proposed changes",
    )
    await proposed_change_permission.save(db=db)

    # Other permissions, created to keep references of them from the start
    for permission_action, permission_description in (
        (GlobalPermissions.EDIT_DEFAULT_BRANCH, "Allow a user to change data in the default branch"),
        (GlobalPermissions.MANAGE_ACCOUNTS, "Allow a user to manage accounts, account roles and account groups"),
        (GlobalPermissions.MANAGE_PERMISSIONS, "Allow a user to manage permissions"),
        (GlobalPermissions.MERGE_BRANCH, "Allow a user to merge branches"),
    ):
        permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
        await permission.new(
            db=db,
            action=permission_action.value,
            decision=PermissionDecision.ALLOW_ALL.value,
            description=permission_description,
        )
        await permission.save(db=db)

    view_permission = await Node.init(db=db, schema=InfrahubKind.OBJECTPERMISSION)
    await view_permission.new(
        db=db,
        name="*",
        namespace="*",
        action=PermissionAction.VIEW.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to view any object in any branch",
    )
    await view_permission.save(db=db)

    modify_permission = await Node.init(db=db, schema=InfrahubKind.OBJECTPERMISSION)
    await modify_permission.new(
        db=db,
        name="*",
        namespace="*",
        action=PermissionAction.ANY.value,
        decision=PermissionDecision.ALLOW_OTHER.value,
        description="Allow a user to change data in non-default branches",
    )
    await modify_permission.save(db=db)

    role_name = "General Access"
    role = await Node.init(db=db, schema=InfrahubKind.ACCOUNTROLE)
    await role.new(
        db=db,
        name=role_name,
        permissions=[
            repo_permission,
            schema_permission,
            proposed_change_permission,
            view_permission,
            modify_permission,
        ],
    )
    await role.save(db=db)
    log.info(f"Created account role: {role_name}")

    group_name = "Infrahub Users"
    group = await Node.init(db=db, schema=InfrahubKind.ACCOUNTGROUP)
    await group.new(db=db, name=group_name, roles=[role])
    await group.save(db=db)
    log.info(f"Created account group: {group_name}")

    return role


async def create_anonymous_role(db: InfrahubDatabase) -> Node:
    deny_permission = await Node.init(db=db, schema=InfrahubKind.OBJECTPERMISSION)
    await deny_permission.new(
        db=db, name="*", namespace="*", action=PermissionAction.ANY.value, decision=PermissionDecision.DENY.value
    )
    await deny_permission.save(db=db)

    view_permission = await NodeManager.get_one_by_hfid(
        db=db,
        kind=InfrahubKind.OBJECTPERMISSION,
        hfid=["*", "*", PermissionAction.VIEW.value, str(PermissionDecision.ALLOW_ALL.value)],
    )

    role = await Node.init(db=db, schema=InfrahubKind.ACCOUNTROLE)
    await role.new(
        db=db, name=config.SETTINGS.main.anonymous_access_role, permissions=[deny_permission, view_permission]
    )
    await role.save(db=db)
    log.info(f"Created anonymous account role: {config.SETTINGS.main.anonymous_access_role}")

    return role


async def create_super_administrators_group(
    db: InfrahubDatabase, role: Node, admin_accounts: list[CoreAccount]
) -> Node:
    group_name = "Super Administrators"
    group = await Node.init(db=db, schema=InfrahubKind.ACCOUNTGROUP)
    await group.new(db=db, name=group_name, roles=[role])
    await group.save(db=db)
    log.info(f"Created account group: {group_name}")

    for admin_account in admin_accounts:
        await group.members.add(db=db, data=admin_account)  # type: ignore[attr-defined]
        await group.members.save(db=db)  # type: ignore[attr-defined]
        log.info(f"Assigned account group: {group_name} to {admin_account.name.value}")

    return group


async def first_time_initialization(db: InfrahubDatabase) -> None:
    # --------------------------------------------------
    # Create the default Branch
    # --------------------------------------------------
    await create_root_node(db=db)
    default_branch = await create_default_branch(db=db)
    await create_global_branch(db=db)

    # --------------------------------------------------
    # Load the internal and core schema in the database
    # --------------------------------------------------
    registry.schema = SchemaManager()
    schema = SchemaRoot(**internal_schema)
    schema_branch = registry.schema.register_schema(schema=schema, branch=default_branch.name)
    schema_branch.load_schema(schema=SchemaRoot(**core_models))
    schema_branch.process()
    await registry.schema.load_schema_to_db(schema=schema_branch, branch=default_branch, db=db)
    registry.schema.set_schema_branch(name=default_branch.name, schema=schema_branch)
    default_branch.update_schema_hash()
    await default_branch.save(db=db)
    log.info("Created the Schema in the database", hash=default_branch.active_schema_hash.main)

    # --------------------------------------------------
    # Create Default Menu
    # --------------------------------------------------
    await create_default_menu(db=db)

    # --------------------------------------------------
    # Create Default Users and Groups
    # --------------------------------------------------
    admin_accounts: list[CoreAccount] = []
    admin_accounts.append(
        await create_account(
            db=db,
            name="admin",
            password=config.SETTINGS.initial.admin_password,
            token_value=config.SETTINGS.initial.admin_token,
        )
    )

    if config.SETTINGS.initial.create_agent_user:
        password = config.SETTINGS.initial.agent_password or str(uuid4())

        admin_accounts.append(
            await create_account(
                db=db, name="agent", password=password, token_value=config.SETTINGS.initial.agent_token
            )
        )

    # --------------------------------------------------
    # Create Global Permissions and assign them
    # --------------------------------------------------
    administrator_role = await create_super_administrator_role(db=db)
    await create_super_administrators_group(db=db, role=administrator_role, admin_accounts=admin_accounts)

    await create_default_roles(db=db)
    if config.SETTINGS.main.allow_anonymous_access:
        await create_anonymous_role(db=db)

    # --------------------------------------------------
    # Create Default IPAM Namespace
    # --------------------------------------------------
    await create_ipam_namespace(db=db)



from __future__ import annotations

from typing import Any

from jinja2 import TemplateSyntaxError, meta, nodes
from jinja2.sandbox import SandboxedEnvironment

ALLOWED_FILTERS = [
    "abs",
    "capitalize",
    "center",
    "default",
    "escape",
    "filesizeformat",
    "join",
    "length",
    "list",
    "lower",
    "replace",
    "reverse",
    "round",
    "sort",
    "sum",
    "title",
    "trim",
    "truncate",
    "upper",
    "wordcount",
]


class MacroDefinition:
    def __init__(self, macro: str) -> None:
        self.macro = macro
        self.env = SandboxedEnvironment()
        self._filters: list[str] = []
        self._variables: list[str] = []
        self.template = self._parse_template()

    def _parse_template(self) -> nodes.Template:
        try:
            self.env.globals = {}
            template = self.env.parse(self.macro)
            if any(node.__class__.__name__ in ["Call", "Import", "Include"] for node in template.body):
                raise ValueError("Forbidden code found in the macro")

        except TemplateSyntaxError as exc:
            raise ValueError(f"Failed to parse the macro: {exc}") from exc

        for node in template.find_all(nodes.Filter):
            if node.name not in ALLOWED_FILTERS:
                raise ValueError(f"The '{node.name}' filter isn't allowed to be used")
            self._filters.append(node.name)

        self._variables = list(meta.find_undeclared_variables(template))

        return template

    def render(self, variables: dict[str, Any]) -> str:
        template = self.env.from_string(self.macro)
        return template.render(variables)

    @property
    def filters(self) -> list[str]:
        return self._filters

    @property
    def variables(self) -> list[str]:
        return self._variables



from __future__ import annotations

from dataclasses import dataclass

from infrahub.database import InfrahubDatabase, get_db


@dataclass
class CliContext:
    application: str = "infrahub.server:app"

    # This method is inherited for Infrahub Enterprise.
    @staticmethod
    async def init_db(retry: int) -> InfrahubDatabase:
        return InfrahubDatabase(driver=await get_db(retry=retry))



import base64
import hashlib
import hmac
from datetime import datetime, timezone
from math import floor
from typing import Any, Optional, Union
from uuid import uuid4

from infrahub_sdk.protocols import CoreTransformPython
from pydantic import BaseModel, ConfigDict, Field

from infrahub.core.constants import InfrahubKind
from infrahub.git.repository import InfrahubReadOnlyRepository, InfrahubRepository
from infrahub.services import InfrahubServices
from infrahub.transformations.constants import DEFAULT_TRANSFORM_TIMEOUT


class SendWebhookData(BaseModel):
    """Sent a webhook to an external source."""

    webhook_id: str = Field(..., description="The unique ID of the webhook")
    event_type: str = Field(..., description="The event type")
    event_data: dict = Field(..., description="The data tied to the event")


class Webhook(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    service: InfrahubServices = Field(...)
    url: str = Field(...)
    event_type: str = Field(...)
    data: dict[str, Any] = Field(...)
    validate_certificates: bool = Field(...)
    _payload: Any = None
    _headers: Optional[dict[str, Any]] = None

    async def _prepare_payload(self) -> None:
        self._payload = {"event_type": self.event_type, "data": self.data}

    def _assign_headers(self) -> None:
        self._headers = {}

    @property
    def webhook_type(self) -> str:
        return self.__class__.__name__

    async def send(self) -> None:
        await self._prepare_payload()
        self._assign_headers()
        await self.service.http.post(url=self.url, json=self._payload, headers=self._headers)


class CustomWebhook(Webhook):
    """Custom webhook"""


class StandardWebhook(Webhook):
    shared_key: bytes = Field(...)

    def _assign_headers(self) -> None:
        message_id = f"msg_{uuid4().hex}"
        timestamp = str(floor(datetime.now(tz=timezone.utc).timestamp()))
        payload = self._payload or {}
        unsigned_data = f"{message_id}.{timestamp}.{payload}".encode()
        signature = self._sign(data=unsigned_data)

        self._headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "webhook-id": message_id,
            "webhook-timestamp": timestamp,
            "webhook-signature": f"v1,{base64.b64encode(signature).decode('utf-8')}",
        }

    def _sign(self, data: bytes) -> bytes:
        return hmac.new(key=self.shared_key, msg=data, digestmod=hashlib.sha256).digest()


class TransformWebhook(Webhook):
    repository_id: str = Field(...)
    repository_name: str = Field(...)
    repository_kind: str = Field(...)
    transform_name: str = Field(...)
    transform_class: str = Field(...)
    transform_file: str = Field(...)

    async def _prepare_payload(self) -> None:
        repo: Union[InfrahubReadOnlyRepository, InfrahubRepository]
        if self.repository_kind == InfrahubKind.READONLYREPOSITORY:
            repo = await InfrahubReadOnlyRepository.init(
                id=self.repository_id, name=self.repository_name, client=self.service.client
            )
        else:
            repo = await InfrahubRepository.init(
                id=self.repository_id, name=self.repository_name, client=self.service.client
            )

        default_branch = repo.default_branch
        commit = repo.get_commit_value(branch_name=default_branch)

        timeout = DEFAULT_TRANSFORM_TIMEOUT
        if transform := await self.service.client.get(
            kind=CoreTransformPython, name__value=self.transform_name, raise_when_missing=False
        ):
            timeout = transform.timeout.value

        self._payload = await repo.execute_python_transform.with_options(timeout_seconds=timeout)(
            branch_name=default_branch,
            commit=commit,
            location=f"{self.transform_file}::{self.transform_class}",
            data={"event_type": self.event_type, "data": self.data},
            client=self.service.client,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class Migration002Query01(Query):
    name = "migration_002_01"
    type: QueryType = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        query = """
        MATCH (a:AttributeValue)
        WHERE a.is_default IS NULL
        SET a.is_default = false
        """
        self.add_to_query(query)
        self.return_labels = ["a"]


class Migration002(GraphMigration):
    name: str = "002_attribute_is_default"
    queries: Sequence[type[Query]] = [Migration002Query01]
    minimum_version: int = 1

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()

        return result



from __future__ import annotations

import re
from enum import Enum

from infrahub_sdk.diff import NodeDiff  # noqa: TC002
from pydantic import BaseModel, Field

from infrahub.core.constants import InfrahubKind, RepositoryInternalStatus
from infrahub.exceptions import NodeNotFoundError

SCHEMA_CHANGE = re.compile("^Schema[A-Z]")


class MessageTTL(int, Enum):
    """Defines the message TTL in seconds, the values themselves are in milliseconds."""

    FIVE = 5000
    TEN = 10000
    TWENTY = 20000

    @classmethod
    def variations(cls) -> list[MessageTTL]:
        """Return available variations of message time to live."""
        return [cls(cls.__members__[member].value) for member in list(cls.__members__)]


class KVTTL(int, Enum):
    """Defines the KV TTL in seconds."""

    ONE = 1
    TEN = 10
    FIFTEEN = 15
    TWO_HOURS = 7200

    @classmethod
    def variations(cls) -> list[KVTTL]:
        """Return available variations of KV time to live."""
        return [cls(cls.__members__[member].value) for member in list(cls.__members__)]


class ProposedChangeRepository(BaseModel):
    repository_id: str
    repository_name: str
    read_only: bool
    source_branch: str
    destination_branch: str
    internal_status: str
    source_commit: str = Field(default="")
    destination_commit: str = Field(default="")
    conflicts: list[str] = Field(default_factory=list, description="List of files with merge conflicts")
    files_added: list[str] = Field(default_factory=list)
    files_changed: list[str] = Field(default_factory=list)
    files_removed: list[str] = Field(default_factory=list)

    @property
    def has_diff(self) -> bool:
        """Indicates if a diff exists for managed repositories."""
        if not self.read_only and self.source_commit and self.source_commit != self.destination_commit:
            return True
        return False

    @property
    def is_staging(self) -> bool:
        """Indicates if the repository is in staging mode."""
        if self.internal_status == RepositoryInternalStatus.STAGING.value:
            return True
        return False

    @property
    def kind(self) -> str:
        if self.read_only:
            return InfrahubKind.READONLYREPOSITORY
        return InfrahubKind.REPOSITORY

    @property
    def has_modifications(self) -> bool:
        """Indicates if any of the files in the repository has been modified."""
        return bool(self.files_added + self.files_changed + self.files_removed)


class ProposedChangeSubscriber(BaseModel):
    subscriber_id: str
    kind: str


class ProposedChangeArtifactDefinition(BaseModel):
    definition_id: str
    definition_name: str
    artifact_name: str
    query_name: str
    query_models: list[str]
    repository_id: str
    transform_kind: str
    template_path: str = Field(default="")
    class_name: str = Field(default="")
    content_type: str
    file_path: str = Field(default="")
    timeout: int

    @property
    def transform_location(self) -> str:
        if self.transform_kind == InfrahubKind.TRANSFORMJINJA2:
            return self.template_path
        if self.transform_kind == InfrahubKind.TRANSFORMPYTHON:
            return f"{self.file_path}::{self.class_name}"

        raise ValueError("Invalid kind for Transform")


class ProposedChangeBranchDiff(BaseModel):
    diff_summary: list[NodeDiff] = Field(default_factory=list, description="The DiffSummary between two branches")
    repositories: list[ProposedChangeRepository] = Field(default_factory=list)
    subscribers: list[ProposedChangeSubscriber] = Field(default_factory=list)

    def get_repository(self, repository_id: str) -> ProposedChangeRepository:
        for repository in self.repositories:
            if repository_id == repository.repository_id:
                return repository
        raise NodeNotFoundError(node_type="Repository", identifier=repository_id)

    def get_subscribers_ids(self, kind: str) -> list[str]:
        return [subscriber.subscriber_id for subscriber in self.subscribers if subscriber.kind == kind]

    def has_node_changes(self, branch: str) -> bool:
        """Indicates if there is at least one node object that has been modified in the branch"""
        return bool(
            [
                entry
                for entry in self.diff_summary
                if entry["branch"] == branch and not SCHEMA_CHANGE.match(entry["kind"])
            ]
        )

    def has_data_changes(self, branch: str) -> bool:
        """Indicates if there are node or schema changes within the branch."""
        return bool([entry for entry in self.diff_summary if entry["branch"] == branch])

    @property
    def has_file_modifications(self) -> bool:
        """Indicates modifications to any of the files in the Git repositories."""
        return any(repository.has_modifications for repository in self.repositories)

    def modified_nodes(self, branch: str) -> list[str]:
        """Return a list of non schema nodes that have been modified on the branch"""
        return [
            entry["id"]
            for entry in self.diff_summary
            if entry["branch"] == branch and not SCHEMA_CHANGE.match(entry["kind"])
        ]

    def modified_kinds(self, branch: str) -> list[str]:
        """Return a list of non schema kinds that have been modified on the branch"""
        return list(
            {
                entry["kind"]
                for entry in self.diff_summary
                if entry["branch"] == branch and not SCHEMA_CHANGE.match(entry["kind"])
            }
        )



from infrahub.core.diff.enricher.hierarchy import DiffHierarchyEnricher
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from ..parent_node_adder import DiffParentNodeAdderDependency


class DiffHierarchyEnricherDependency(DependencyBuilder[DiffHierarchyEnricher]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffHierarchyEnricher:
        return DiffHierarchyEnricher(db=context.db, parent_adder=DiffParentNodeAdderDependency.build(context=context))



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from infrahub.core.constants import BranchSupportType, RelationshipStatus
from infrahub.core.graph.schema import GraphRelationshipRelationships, GraphRelDirection
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class SchemaRelationshipInfo(BaseModel):
    name: str
    branch_support: str = BranchSupportType.AWARE.value
    src_peer: str
    dst_peer: str


class RelationshipDuplicateQuery(Query):
    name = "relationship_duplicate"
    type = QueryType.WRITE
    insert_return: bool = False

    def __init__(
        self,
        previous_rel: SchemaRelationshipInfo,
        new_rel: SchemaRelationshipInfo,
        **kwargs: Any,
    ) -> None:
        self.previous_rel = previous_rel
        self.new_rel = new_rel

        super().__init__(**kwargs)

    def render_match(self) -> str:
        query = """
        // Find all the active nodes
        MATCH (source:%(src_peer)s)-[:IS_RELATED]-(rel:Relationship { name: $previous_rel.name })-[:IS_RELATED]-(destination:%(dst_peer)s)
        WHERE source <> destination
        """ % {"src_peer": self.previous_rel.src_peer, "dst_peer": self.previous_rel.dst_peer}

        return query

    @staticmethod
    def _render_sub_query_per_rel_type(rel_name: str, rel_type: str, direction: GraphRelDirection) -> str:
        subquery = [
            f"WITH peer_node, {rel_name}, active_rel, new_rel",
            f"WITH peer_node, {rel_name}, active_rel, new_rel",
            f'WHERE type({rel_name}) = "{rel_type}"',
        ]
        if direction == GraphRelDirection.OUTBOUND:
            subquery.append(f"CREATE (new_rel)-[:{rel_type} $rel_props_new ]->(peer_node)")
            subquery.append(f"CREATE (active_rel)-[:{rel_type} $rel_props_prev ]->(peer_node)")
        elif direction == GraphRelDirection.INBOUND:
            subquery.append(f"CREATE (new_rel)<-[:{rel_type} $rel_props_new ]-(peer_node)")
            subquery.append(f"CREATE (active_rel)<-[:{rel_type} $rel_props_prev ]-(peer_node)")
        subquery.append("RETURN peer_node as p2")
        return "\n".join(subquery)

    @classmethod
    def _render_sub_query_out(cls) -> str:
        sub_queries_out = [
            cls._render_sub_query_per_rel_type(
                rel_name="rel_outband", rel_type=rel_type, direction=GraphRelDirection.OUTBOUND
            )
            for rel_type, rel_def in GraphRelationshipRelationships.model_fields.items()
            if rel_def.default.direction in [GraphRelDirection.OUTBOUND, GraphRelDirection.EITHER]
        ]
        sub_query_out = "\nUNION\n".join(sub_queries_out)
        return sub_query_out

    @classmethod
    def _render_sub_query_in(cls) -> str:
        sub_queries_in = [
            cls._render_sub_query_per_rel_type(
                rel_name="rel_inband", rel_type=rel_type, direction=GraphRelDirection.INBOUND
            )
            for rel_type, rel_def in GraphRelationshipRelationships.model_fields.items()
            if rel_def.default.direction in [GraphRelDirection.INBOUND, GraphRelDirection.EITHER]
        ]
        sub_query_in = "\nUNION\n".join(sub_queries_in)
        return sub_query_in

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["new_rel"] = self.new_rel.model_dump()
        self.params["previous_rel"] = self.previous_rel.model_dump()

        self.params["current_time"] = self.at.to_string()
        self.params["branch_name"] = self.branch.name
        self.params["branch_support"] = self.new_rel.branch_support

        self.params["rel_props_new"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
        }

        self.params["rel_props_prev"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
        }

        sub_query_out = self._render_sub_query_out()
        sub_query_in = self._render_sub_query_in()

        self.add_to_query(self.render_match())

        # ruff: noqa: E501
        query = """
        CALL {
            WITH source, rel, destination
            MATCH path = (source)-[r1:IS_RELATED]-(rel)-[r2:IS_RELATED]-(destination)
            WHERE all(r IN relationships(path) WHERE %(branch_filter)s)
            RETURN rel as rel1, r1 as r11, r2 as r12
            ORDER BY r1.branch_level DESC, r2.branch_level DESC, r1.from DESC, r2.from DESC
            LIMIT 1
        }
        WITH rel1 as active_rel, r11 as r1, r12 as r2
        WHERE r1.status = "active" AND r2.status = "active"
        CREATE (new_rel:Relationship { uuid: active_rel.uuid, name: $new_rel.name, branch_support: $new_rel.branch_support })
        WITH DISTINCT(active_rel) as active_rel, new_rel
        // Process Inbound Relationship
        MATCH (active_rel)<-[]-(peer)
        CALL {
            WITH active_rel, peer
            MATCH (active_rel)<-[r]-(peer)
            WHERE %(branch_filter)s
            RETURN active_rel as n1, r as rel_inband1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_rel, rel_inband1 as rel_inband, p1 as peer_node, new_rel
        WHERE rel_inband.status = "active"
        CALL {
            %(sub_query_in)s
        }
        WITH p2 as peer_node, rel_inband, active_rel, new_rel
        FOREACH (i in CASE WHEN rel_inband.branch = $branch_name THEN [1] ELSE [] END |
            SET rel_inband.to = $current_time
        )
        WITH DISTINCT(active_rel) as active_rel, new_rel
        // Process Outbound Relationship
        MATCH (active_rel)-[]->(peer)
        CALL {
            WITH active_rel, peer
            MATCH (active_rel)-[r]->(peer)
            WHERE %(branch_filter)s
            RETURN active_rel as n1, r as rel_outband1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_rel, rel_outband1 as rel_outband, p1 as peer_node, new_rel
        WHERE rel_outband.status = "active"
        CALL {
            %(sub_query_out)s
        }
        WITH p2 as peer_node, rel_outband, active_rel, new_rel
        FOREACH (i in CASE WHEN rel_outband.branch = $branch_name THEN [1] ELSE [] END |
            SET rel_outband.to = $current_time
        )
        RETURN DISTINCT new_rel
        """ % {
            "branch_filter": branch_filter,
            "sub_query_out": sub_query_out,
            "sub_query_in": sub_query_in,
        }
        self.add_to_query(query)



from infrahub import config
from infrahub.auth import AccountSession
from infrahub.core import registry
from infrahub.core.account import GlobalPermission
from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME, GlobalPermissions, PermissionDecision
from infrahub.database import InfrahubDatabase
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.initialization import GraphqlParams

from .interface import CheckerResolution, GraphQLQueryPermissionCheckerInterface


class DefaultBranchPermissionChecker(GraphQLQueryPermissionCheckerInterface):
    """Checker that makes sure a user account can edit data in the default branch."""

    permission_required = GlobalPermission(
        action=GlobalPermissions.EDIT_DEFAULT_BRANCH.value, decision=PermissionDecision.ALLOW_ALL.value
    )
    exempt_operations = [
        "BranchCreate",
        "DiffUpdate",
        "InfrahubAccountSelfUpdate",
        "InfrahubAccountTokenCreate",
        "InfrahubAccountTokenDelete",
    ]

    async def supports(self, db: InfrahubDatabase, account_session: AccountSession, branch: Branch) -> bool:
        return config.SETTINGS.main.allow_anonymous_access or account_session.authenticated

    async def check(
        self,
        db: InfrahubDatabase,
        account_session: AccountSession,
        analyzed_query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,
        branch: Branch,
    ) -> CheckerResolution:
        operates_on_default_branch = analyzed_query.branch is None or analyzed_query.branch.name in (
            GLOBAL_BRANCH_NAME,
            registry.default_branch,
        )
        is_exempt_operation = all(
            operation_name in self.exempt_operations for operation_name in analyzed_query.operation_names
        )

        if operates_on_default_branch and analyzed_query.contains_mutation and not is_exempt_operation:
            query_parameters.context.active_permissions.raise_for_permission(permission=self.permission_required)

        return CheckerResolution.NEXT_CHECKER



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Optional
from uuid import uuid4

from infrahub.core.constants import (
    NULL_VALUE,
    BranchSupportType,
    DiffAction,
    RelationshipCardinality,
    RelationshipStatus,
)
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.timestamp import Timestamp

from .model.path import (
    DatabasePath,
    DiffAttribute,
    DiffNode,
    DiffProperty,
    DiffRelationship,
    DiffRoot,
    DiffSingleRelationship,
)

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.query import QueryResult
    from infrahub.core.schema.manager import SchemaManager
    from infrahub.core.schema.relationship_schema import RelationshipSchema


class DiffNoChildPathError(Exception): ...


class DiffNoPeerIdError(Exception): ...


class DiffNotFoundError(Exception): ...


@dataclass
class TrackedStatusUpdates:
    timestamp_status_map: dict[Timestamp, RelationshipStatus] = field(kw_only=True, default_factory=dict)
    _ordered_timestamps: list[Timestamp] = field(kw_only=True, default_factory=list)

    def track_database_path(self, database_path: DatabasePath) -> None:
        if database_path.node_changed_at in self.timestamp_status_map:
            return
        self.timestamp_status_map[database_path.node_changed_at] = database_path.node_status

    def get_ordered_timestamps_asc(self) -> list[Timestamp]:
        if self._ordered_timestamps:
            return self._ordered_timestamps
        if not self.timestamp_status_map:
            return []
        self._ordered_timestamps = sorted(self.timestamp_status_map.keys())
        return self._ordered_timestamps

    def get_action_and_timestamp(self, from_time: Timestamp) -> tuple[DiffAction, Timestamp]:
        ordered_timestamps = self.get_ordered_timestamps_asc()
        if len(ordered_timestamps) == 0:
            raise DiffNoChildPathError()
        latest_timestamp = ordered_timestamps[-1]
        if latest_timestamp < from_time:
            return (DiffAction.UPDATED, latest_timestamp)
        latest_status = self.timestamp_status_map[latest_timestamp]
        action = DiffAction.REMOVED if latest_status is RelationshipStatus.DELETED else DiffAction.ADDED
        return (action, latest_timestamp)


@dataclass
class DiffValueIntermediate:
    changed_at: Timestamp
    status: RelationshipStatus
    value: Any

    def __hash__(self) -> int:
        return hash("|".join(str(v) for v in (self.changed_at.to_string(), self.status.value, self.value)))


@dataclass
class DiffPropertyIntermediate:
    property_type: DatabaseEdgeType
    diff_values: set[DiffValueIntermediate] = field(default_factory=set)
    _ordered_values: list[DiffValueIntermediate] = field(default_factory=list)

    def add_value(self, diff_value: DiffValueIntermediate) -> None:
        self.diff_values.add(diff_value)

    def get_ordered_values_asc(self) -> list[DiffValueIntermediate]:
        if self._ordered_values:
            return self._ordered_values
        if not self.diff_values:
            return []
        self._ordered_values = sorted(self.diff_values, key=lambda df: df.changed_at)
        return self._ordered_values

    @property
    def earliest_diff_value(self) -> Optional[DiffValueIntermediate]:
        ordered_values = self.get_ordered_values_asc()
        if not ordered_values:
            return None
        return ordered_values[0]

    def get_property_details(self, from_time: Timestamp) -> tuple[DiffAction, Timestamp, Any, Any]:
        """Returns action, timestamp, previous_value, new_value"""
        ordered_values = self.get_ordered_values_asc()
        previous: Any = None
        new: Any = None
        if len(ordered_values) == 0:
            raise DiffNoChildPathError()
        if len(ordered_values) == 1:
            lone_value = ordered_values[0]
            if lone_value.changed_at < from_time:
                action = DiffAction.UNCHANGED
                previous = new = lone_value.value
            elif lone_value.status is RelationshipStatus.ACTIVE:
                action = DiffAction.ADDED
                new = lone_value.value
            else:
                action = DiffAction.REMOVED
                previous = lone_value.value
            return (action, lone_value.changed_at, previous, new)
        earliest_value = ordered_values[0]
        new_diff_value = ordered_values[-1]
        new_value = new_diff_value.value
        new_status = new_diff_value.status
        action = DiffAction.UPDATED
        previous_value = earliest_value.value
        if earliest_value.changed_at > from_time:
            action = DiffAction.ADDED
            previous_value = None
        elif earliest_value.value in (None, NULL_VALUE) and new_value not in (None, NULL_VALUE):
            action = DiffAction.ADDED
        elif (earliest_value.value not in (None, NULL_VALUE) and new_value in (None, NULL_VALUE)) or (
            earliest_value.status,
            new_status,
        ) == (RelationshipStatus.ACTIVE, RelationshipStatus.DELETED):
            action = DiffAction.REMOVED
            if new_value != NULL_VALUE:
                new_value = None
        elif earliest_value.value == new_value or {earliest_value.value, new_value} <= {None, NULL_VALUE}:
            action = DiffAction.UNCHANGED
        return (action, new_diff_value.changed_at, previous_value, new_value)

    def to_diff_property(self, from_time: Timestamp) -> DiffProperty:
        action, changed_at, previous_value, new_value = self.get_property_details(from_time=from_time)
        return DiffProperty(
            property_type=self.property_type,
            changed_at=changed_at,
            action=action,
            previous_value=previous_value,
            new_value=new_value,
        )


@dataclass
class DiffAttributeIntermediate(TrackedStatusUpdates):
    uuid: str
    name: str
    from_time: Timestamp
    properties_by_type: dict[DatabaseEdgeType, DiffPropertyIntermediate] = field(default_factory=dict)

    def track_database_path(self, database_path: DatabasePath) -> None:
        if database_path.attribute_changed_at in self.timestamp_status_map:
            return
        self.timestamp_status_map[database_path.attribute_changed_at] = database_path.attribute_status

    def to_diff_attribute(self, include_unchanged: bool) -> DiffAttribute:
        properties = []
        for prop in self.properties_by_type.values():
            diff_prop = prop.to_diff_property(from_time=self.from_time)
            if include_unchanged or diff_prop.action is not DiffAction.UNCHANGED:
                properties.append(diff_prop)
        action, changed_at = self.get_action_and_timestamp(from_time=self.from_time)
        if not properties or all(p.action is DiffAction.UNCHANGED for p in properties):
            action = DiffAction.UNCHANGED
        return DiffAttribute(
            uuid=self.uuid, name=self.name, changed_at=changed_at, action=action, properties=properties
        )


@dataclass
class DiffRelationshipPropertyIntermediate:
    property_type: DatabaseEdgeType
    db_id: str
    changed_at: Timestamp
    status: RelationshipStatus
    value: Any

    def __hash__(self) -> int:
        return hash(
            "|".join(
                (self.property_type.value, self.db_id, self.changed_at.to_string(), self.status.value, str(self.value))
            )
        )


@dataclass
class DiffSingleRelationshipIntermediate:
    peer_id: str
    last_changed_at: Timestamp
    ordered_properties_by_type: dict[DatabaseEdgeType, list[DiffRelationshipPropertyIntermediate]] = field(
        default_factory=dict
    )

    @classmethod
    def from_properties(
        cls, properties: list[DiffRelationshipPropertyIntermediate]
    ) -> DiffSingleRelationshipIntermediate:
        if not properties:
            raise DiffNoChildPathError()
        peer_id = None
        for diff_property in properties:
            if diff_property.property_type == DatabaseEdgeType.IS_RELATED:
                peer_id = diff_property.value
                break
        if not peer_id:
            raise DiffNoPeerIdError(f"Cannot identify peer ID for relationship property {(properties[0]).db_id}")

        ordered_properties_by_type: dict[DatabaseEdgeType, list[DiffRelationshipPropertyIntermediate]] = {}
        # tiebreaker for simultaneous updates is to prefer the DELETED relationship
        chronological_properties = sorted(
            properties, key=lambda p: (p.changed_at, p.status is RelationshipStatus.ACTIVE)
        )
        last_changed_at = chronological_properties[-1].changed_at
        for chronological_property in chronological_properties:
            property_key = DatabaseEdgeType(chronological_property.property_type)
            if chronological_property.property_type not in ordered_properties_by_type:
                ordered_properties_by_type[property_key] = []
            ordered_properties_by_type[property_key].append(chronological_property)
        return cls(
            peer_id=peer_id,
            last_changed_at=last_changed_at,
            ordered_properties_by_type=ordered_properties_by_type,
        )

    def _get_single_relationship_final_property(
        self, chronological_properties: list[DiffRelationshipPropertyIntermediate], from_time: Timestamp
    ) -> DiffProperty:
        property_type = DatabaseEdgeType(chronological_properties[0].property_type)
        changed_at = chronological_properties[-1].changed_at
        previous_value = None
        first_diff_prop = chronological_properties[0]
        if first_diff_prop.status is RelationshipStatus.DELETED or first_diff_prop.changed_at < from_time:
            previous_value = first_diff_prop.value
        new_value = None
        last_diff_prop = chronological_properties[-1]
        changed_at = last_diff_prop.changed_at
        if last_diff_prop is first_diff_prop:
            if last_diff_prop.status is RelationshipStatus.DELETED:
                previous_value = last_diff_prop.value
            else:
                new_value = last_diff_prop.value
        elif last_diff_prop.status != RelationshipStatus.DELETED:
            new_value = last_diff_prop.value
        action = DiffAction.UPDATED
        if last_diff_prop.changed_at < from_time:
            action = DiffAction.UNCHANGED
        elif previous_value in (None, NULL_VALUE) and new_value not in (None, NULL_VALUE):
            action = DiffAction.ADDED
        elif previous_value not in (None, NULL_VALUE) and new_value in (None, NULL_VALUE):
            action = DiffAction.REMOVED
        elif previous_value == new_value or {previous_value, new_value} <= {None, NULL_VALUE}:
            action = DiffAction.UNCHANGED
        return DiffProperty(
            property_type=property_type,
            changed_at=changed_at,
            previous_value=previous_value,
            new_value=new_value,
            action=action,
        )

    def get_final_single_relationship(self, from_time: Timestamp, include_unchanged: bool) -> DiffSingleRelationship:
        final_properties = []
        peer_id_properties = self.ordered_properties_by_type[DatabaseEdgeType.IS_RELATED]
        peer_final_property = self._get_single_relationship_final_property(
            chronological_properties=peer_id_properties, from_time=from_time
        )
        peer_id = peer_final_property.new_value or peer_final_property.previous_value
        other_final_properties = []
        for property_type, property_list in self.ordered_properties_by_type.items():
            if property_type is DatabaseEdgeType.IS_RELATED:
                continue
            final_prop = self._get_single_relationship_final_property(
                chronological_properties=property_list, from_time=from_time
            )
            if include_unchanged or final_prop.action is not DiffAction.UNCHANGED:
                other_final_properties.append(final_prop)
        final_properties = [peer_final_property] + other_final_properties
        last_changed_property = max(final_properties, key=lambda fp: fp.changed_at)
        last_changed_at = last_changed_property.changed_at
        # handle case when peer has been deleted on another branch, but the properties of the relationship are updated
        if (
            last_changed_property.property_type is not DatabaseEdgeType.IS_RELATED
            and any(p.action is not DiffAction.REMOVED for p in other_final_properties)
            and peer_final_property.action is DiffAction.REMOVED
        ):
            peer_final_property.action = DiffAction.UNCHANGED
            peer_final_property.new_value = peer_id
        if last_changed_at < from_time or all(fp.action is DiffAction.UNCHANGED for fp in final_properties):
            action = DiffAction.UNCHANGED
        elif peer_final_property.action in (DiffAction.ADDED, DiffAction.REMOVED):
            action = peer_final_property.action
        else:
            action = DiffAction.UPDATED
        return DiffSingleRelationship(
            peer_id=peer_id, changed_at=last_changed_at, action=action, properties=final_properties
        )


@dataclass
class DiffRelationshipIntermediate:
    name: str
    identifier: str
    cardinality: RelationshipCardinality
    from_time: Timestamp
    properties_by_db_id: dict[str, set[DiffRelationshipPropertyIntermediate]] = field(default_factory=dict)
    _single_relationship_list: list[DiffSingleRelationshipIntermediate] = field(default_factory=list)

    def add_path(self, database_path: DatabasePath, diff_from_time: Timestamp, diff_to_time: Timestamp) -> None:
        if database_path.property_type in [
            DatabaseEdgeType.IS_RELATED,
            DatabaseEdgeType.HAS_OWNER,
            DatabaseEdgeType.HAS_SOURCE,
        ]:
            value = database_path.peer_id
        else:
            value = database_path.property_value
        db_id = database_path.attribute_node.element_id
        if db_id not in self.properties_by_db_id:
            self.properties_by_db_id[db_id] = set()
        self.properties_by_db_id[db_id].add(
            DiffRelationshipPropertyIntermediate(
                db_id=db_id,
                property_type=database_path.property_type,
                changed_at=database_path.property_from_time,
                status=database_path.property_status,
                value=value,
            )
        )
        to_time = database_path.property_to_time
        # if the to time was set in the time frame, then it is effectively a delete
        if (
            to_time
            and database_path.property_from_time < diff_from_time <= to_time <= diff_to_time
            and database_path.property_status is RelationshipStatus.ACTIVE
        ):
            self.properties_by_db_id[db_id].add(
                DiffRelationshipPropertyIntermediate(
                    db_id=db_id,
                    property_type=database_path.property_type,
                    changed_at=to_time,
                    status=RelationshipStatus.DELETED,
                    value=value,
                )
            )

    def _index_relationships(self) -> None:
        self._single_relationship_list = [
            DiffSingleRelationshipIntermediate.from_properties(list(single_relationship_properties))
            for single_relationship_properties in self.properties_by_db_id.values()
        ]

    def to_diff_relationship(self, include_unchanged: bool) -> DiffRelationship:
        self._index_relationships()
        single_relationships = [
            sr.get_final_single_relationship(from_time=self.from_time, include_unchanged=include_unchanged)
            for sr in self._single_relationship_list
        ]
        last_changed_relationship = max(single_relationships, key=lambda r: r.changed_at)
        last_changed_at = last_changed_relationship.changed_at
        action = DiffAction.UPDATED
        if last_changed_at < self.from_time or all(sr.action is DiffAction.UNCHANGED for sr in single_relationships):
            action = DiffAction.UNCHANGED
        # bubble up action, excluding UNCHANGED
        if self.cardinality is RelationshipCardinality.ONE:
            actions = {sr.action for sr in single_relationships if sr.action is not DiffAction.UNCHANGED}
            if len(actions) == 1:
                action = actions.pop()
        return DiffRelationship(
            name=self.name,
            identifier=self.identifier,
            changed_at=last_changed_at,
            action=action,
            relationships=single_relationships,
            cardinality=self.cardinality,
        )


@dataclass
class DiffNodeIntermediate(TrackedStatusUpdates):
    force_action: DiffAction | None
    uuid: str
    kind: str
    db_id: str
    from_time: Timestamp
    status: RelationshipStatus
    attributes_by_name: dict[str, DiffAttributeIntermediate] = field(default_factory=dict)
    # {(name, identifier): DiffRelationshipIntermediate}
    relationships_by_identifier: dict[tuple[str, str], DiffRelationshipIntermediate] = field(default_factory=dict)

    def to_diff_node(self, from_time: Timestamp, include_unchanged: bool) -> DiffNode:
        attributes = []
        for attr in self.attributes_by_name.values():
            diff_attr = attr.to_diff_attribute(include_unchanged=include_unchanged)
            if include_unchanged or diff_attr.action is not DiffAction.UNCHANGED:
                attributes.append(diff_attr)
        action, changed_at = self.get_action_and_timestamp(from_time=from_time)
        relationships = []
        for rel in self.relationships_by_identifier.values():
            diff_rel = rel.to_diff_relationship(include_unchanged=include_unchanged)
            if include_unchanged or diff_rel.action is not DiffAction.UNCHANGED:
                relationships.append(diff_rel)
        if not attributes and not relationships:
            action = DiffAction.UNCHANGED
        if all(a.action is DiffAction.UNCHANGED for a in attributes) and all(
            r.action is DiffAction.UNCHANGED for r in relationships
        ):
            action = DiffAction.UNCHANGED
        if self.force_action:
            action = self.force_action
        return DiffNode(
            uuid=self.uuid,
            kind=self.kind,
            changed_at=changed_at,
            action=action,
            attributes=attributes,
            relationships=relationships,
        )

    @property
    def is_empty(self) -> bool:
        return len(self.attributes_by_name) == 0 and len(self.relationships_by_identifier) == 0


@dataclass
class DiffRootIntermediate:
    uuid: str
    branch: str
    nodes_by_id: dict[str, DiffNodeIntermediate] = field(default_factory=dict)

    def to_diff_root(self, from_time: Timestamp, to_time: Timestamp, include_unchanged: bool) -> DiffRoot:
        nodes = []
        for node in self.nodes_by_id.values():
            if node.is_empty:
                continue
            diff_node = node.to_diff_node(from_time=from_time, include_unchanged=include_unchanged)
            if include_unchanged or diff_node.action is not DiffAction.UNCHANGED:
                nodes.append(diff_node)
        return DiffRoot(uuid=self.uuid, branch=self.branch, nodes=nodes, from_time=from_time, to_time=to_time)


class DiffQueryParser:
    def __init__(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        schema_manager: SchemaManager,
        from_time: Timestamp,
        to_time: Optional[Timestamp] = None,
        previous_node_field_specifiers: dict[str, set[str]] | None = None,
    ) -> None:
        self.base_branch_name = base_branch.name
        self.diff_branch_name = diff_branch.name
        self.schema_manager = schema_manager
        self.from_time = from_time
        self.to_time = to_time or Timestamp()
        # if this diff is for the base branch, use from_time b/c create_time would be too much
        if diff_branch.name == base_branch.name:
            self.diff_branched_from_time = from_time
        else:
            self.diff_branched_from_time = Timestamp(diff_branch.get_branched_from())
        self._diff_root_by_branch: dict[str, DiffRootIntermediate] = {}
        self._final_diff_root_by_branch: dict[str, DiffRoot] = {}
        self._previous_node_field_specifiers = previous_node_field_specifiers or {}
        self._new_node_field_specifiers: dict[str, set[str]] | None = None
        self._current_node_field_specifiers: dict[str, set[str]] | None = None

    def get_branches(self) -> set[str]:
        return set(self._final_diff_root_by_branch.keys())

    def get_diff_root_for_branch(self, branch: str) -> DiffRoot:
        if branch not in (self.base_branch_name, self.diff_branch_name):
            raise DiffNotFoundError(f"No diff found for branch {branch}")
        if branch in self._final_diff_root_by_branch:
            return self._final_diff_root_by_branch[branch]
        return DiffRoot(from_time=self.from_time, to_time=self.to_time, uuid=str(uuid4()), branch=branch, nodes=[])

    def get_diff_node_field_specifiers(self) -> dict[str, set[str]]:
        if self.diff_branch_name not in self._diff_root_by_branch:
            return {}
        node_field_specifiers_map: dict[str, set[str]] = defaultdict(set)
        diff_root = self._diff_root_by_branch[self.diff_branch_name]
        for node in diff_root.nodes_by_id.values():
            for attribute_name in node.attributes_by_name:
                node_field_specifiers_map[node.uuid].add(attribute_name)
            for relationship_diff in node.relationships_by_identifier.values():
                node_field_specifiers_map[node.uuid].add(relationship_diff.identifier)
        return node_field_specifiers_map

    def _remove_node_specifiers(
        self, node_specifiers: dict[str, set[str]], node_specifiers_to_remove: dict[str, set[str]]
    ) -> dict[str, set[str]]:
        final_node_specifiers: dict[str, set[str]] = defaultdict(set)
        for node_uuid, field_names_set in node_specifiers.items():
            specifiers_to_remove = node_specifiers_to_remove.get(node_uuid, set())
            final_specifiers = field_names_set - specifiers_to_remove
            if final_specifiers:
                final_node_specifiers[node_uuid] = final_specifiers
        return final_node_specifiers

    def get_new_node_field_specifiers(self) -> dict[str, set[str]]:
        if self._new_node_field_specifiers is not None:
            return self._new_node_field_specifiers
        branch_node_specifiers = self.get_diff_node_field_specifiers()
        new_node_field_specifiers = self._remove_node_specifiers(
            branch_node_specifiers, self._previous_node_field_specifiers
        )
        self._new_node_field_specifiers = new_node_field_specifiers
        return new_node_field_specifiers

    def get_current_node_field_specifiers(self) -> dict[str, set[str]]:
        if self._current_node_field_specifiers is not None:
            return self._current_node_field_specifiers
        new_node_field_specifiers = self.get_new_node_field_specifiers()
        current_node_field_specifiers = self._remove_node_specifiers(
            self._previous_node_field_specifiers, new_node_field_specifiers
        )
        self._current_node_field_specifiers = current_node_field_specifiers
        return current_node_field_specifiers

    def read_result(self, query_result: QueryResult) -> None:
        path = query_result.get_path(label="diff_path")
        database_path = DatabasePath.from_cypher_path(cypher_path=path)
        self._parse_path(database_path=database_path)
        self._current_node_field_specifiers = None
        self._new_node_field_specifiers = None

    def parse(self, include_unchanged: bool = False) -> None:
        self._new_node_field_specifiers = None
        self._current_node_field_specifiers = None
        if len(self._diff_root_by_branch) > 1:
            self._apply_base_branch_previous_values()
            self._remove_empty_base_diff_root()
        self._finalize(include_unchanged=include_unchanged)

    def _parse_path(self, database_path: DatabasePath) -> None:
        to_time = database_path.property_to_time
        # this path was added and removed within the timeframe, so we ignore it
        if to_time and self.from_time <= database_path.property_from_time <= to_time <= self.to_time:
            return

        diff_root = self._get_diff_root(database_path=database_path)
        diff_node = self._get_diff_node(database_path=database_path, diff_root=diff_root)
        self._update_attribute_level(database_path=database_path, diff_node=diff_node)

    def _get_diff_root(self, database_path: DatabasePath) -> DiffRootIntermediate:
        branch = database_path.deepest_branch
        if branch not in self._diff_root_by_branch:
            self._diff_root_by_branch[branch] = DiffRootIntermediate(uuid=str(uuid4()), branch=branch)
        return self._diff_root_by_branch[branch]

    def _get_diff_node(self, database_path: DatabasePath, diff_root: DiffRootIntermediate) -> DiffNodeIntermediate:
        node_id = database_path.node_id
        if node_id not in diff_root.nodes_by_id:
            diff_root.nodes_by_id[node_id] = DiffNodeIntermediate(
                uuid=node_id,
                kind=database_path.node_kind,
                db_id=database_path.node_db_id,
                from_time=database_path.node_changed_at,
                status=database_path.node_status,
                force_action=DiffAction.UPDATED
                if database_path.node_branch_support is BranchSupportType.AGNOSTIC
                else None,
            )
        diff_node = diff_root.nodes_by_id[node_id]
        # special handling for nodes that have their kind updated, which results in 2 nodes with the same uuid
        if diff_node.db_id != database_path.node_db_id and (
            database_path.node_changed_at > diff_node.from_time
            or (
                database_path.node_changed_at >= diff_node.from_time
                and (diff_node.status, database_path.node_status)
                == (RelationshipStatus.DELETED, RelationshipStatus.ACTIVE)
            )
        ):
            diff_node.kind = database_path.node_kind
            diff_node.db_id = database_path.node_db_id
            diff_node.from_time = database_path.node_changed_at
            diff_node.status = database_path.node_status
        diff_node.track_database_path(database_path=database_path)
        return diff_node

    def _get_relationship_schema(self, database_path: DatabasePath) -> RelationshipSchema | None:
        branches_to_check = [database_path.deepest_branch]
        if database_path.deepest_branch == self.diff_branch_name:
            branches_to_check.append(self.base_branch_name)
        for schema_branch_name in branches_to_check:
            node_schema = self.schema_manager.get(
                name=database_path.node_kind, branch=schema_branch_name, duplicate=False
            )
            relationship_schemas = node_schema.get_relationships_by_identifier(id=database_path.attribute_name)
            if len(relationship_schemas) == 1:
                return relationship_schemas[0]
            possible_path_directions = database_path.possible_relationship_directions
            for rel_schema in relationship_schemas:
                if rel_schema.direction in possible_path_directions:
                    return rel_schema
        return None

    def _update_attribute_level(self, database_path: DatabasePath, diff_node: DiffNodeIntermediate) -> None:
        if "Attribute" in database_path.attribute_node.labels:
            diff_attribute = self._get_diff_attribute(database_path=database_path, diff_node=diff_node)
            self._update_attribute_property(database_path=database_path, diff_attribute=diff_attribute)
            return
        relationship_schema = self._get_relationship_schema(database_path=database_path)
        if not relationship_schema:
            return
        diff_relationship = self._get_diff_relationship(
            diff_node=diff_node, relationship_schema=relationship_schema, database_path=database_path
        )
        diff_relationship.add_path(
            database_path=database_path, diff_from_time=self.from_time, diff_to_time=self.to_time
        )

    def _get_diff_attribute(
        self, database_path: DatabasePath, diff_node: DiffNodeIntermediate
    ) -> DiffAttributeIntermediate:
        attribute_name = database_path.attribute_name
        branch_name = database_path.deepest_branch
        from_time = self.from_time
        if branch_name == self.base_branch_name:
            new_node_field_specifiers = self.get_new_node_field_specifiers()
            if attribute_name in new_node_field_specifiers.get(diff_node.uuid, set()):
                from_time = self.diff_branched_from_time
        if attribute_name not in diff_node.attributes_by_name:
            diff_node.attributes_by_name[attribute_name] = DiffAttributeIntermediate(
                uuid=database_path.attribute_id,
                name=attribute_name,
                from_time=from_time,
            )
        diff_attribute = diff_node.attributes_by_name[attribute_name]
        diff_attribute.track_database_path(database_path=database_path)
        return diff_attribute

    def _update_attribute_property(
        self, database_path: DatabasePath, diff_attribute: DiffAttributeIntermediate
    ) -> None:
        property_type = database_path.property_type
        if property_type not in diff_attribute.properties_by_type:
            diff_attribute.properties_by_type[property_type] = DiffPropertyIntermediate(property_type=property_type)
        diff_property = diff_attribute.properties_by_type[property_type]
        value = database_path.property_value
        if database_path.property_is_peer:
            value = database_path.peer_id
        diff_property.add_value(
            diff_value=DiffValueIntermediate(
                changed_at=database_path.property_from_time,
                status=database_path.property_status,
                value=value,
            )
        )

    def _get_diff_relationship(
        self,
        diff_node: DiffNodeIntermediate,
        relationship_schema: RelationshipSchema,
        database_path: DatabasePath,
    ) -> DiffRelationshipIntermediate:
        diff_relationship = diff_node.relationships_by_identifier.get(
            (relationship_schema.name, relationship_schema.get_identifier())
        )
        if not diff_relationship:
            branch_name = database_path.deepest_branch
            from_time = self.from_time
            if branch_name == self.base_branch_name:
                new_node_field_specifiers = self.get_new_node_field_specifiers()
                if relationship_schema.get_identifier() in new_node_field_specifiers.get(diff_node.uuid, set()):
                    from_time = self.diff_branched_from_time
            diff_relationship = DiffRelationshipIntermediate(
                name=relationship_schema.name,
                cardinality=relationship_schema.cardinality,
                identifier=relationship_schema.get_identifier(),
                from_time=from_time,
            )
            diff_node.relationships_by_identifier[relationship_schema.name, relationship_schema.get_identifier()] = (
                diff_relationship
            )
        return diff_relationship

    def _apply_base_branch_previous_values(self) -> None:
        base_diff_root = self._diff_root_by_branch.get(self.base_branch_name)
        if not base_diff_root:
            return
        other_branches = set(self._diff_root_by_branch.keys()) - {self.base_branch_name}
        for branch in other_branches:
            branch_diff_root = self._diff_root_by_branch.get(branch)
            if not branch_diff_root:
                continue
            for node_id, diff_node in branch_diff_root.nodes_by_id.items():
                base_diff_node = base_diff_root.nodes_by_id.get(node_id)
                if not base_diff_node:
                    continue
                self._apply_attribute_previous_values(diff_node=diff_node, base_diff_node=base_diff_node)
                self._apply_relationship_previous_values(diff_node=diff_node, base_diff_node=base_diff_node)

    def _apply_attribute_previous_values(
        self, diff_node: DiffNodeIntermediate, base_diff_node: DiffNodeIntermediate
    ) -> None:
        for attribute_name, diff_attribute in diff_node.attributes_by_name.items():
            base_diff_attribute = base_diff_node.attributes_by_name.get(attribute_name)
            if not base_diff_attribute:
                continue
            for property_type, diff_property in diff_attribute.properties_by_type.items():
                base_diff_property = base_diff_attribute.properties_by_type.get(property_type)
                if not base_diff_property:
                    return
                base_previous_diff_value = base_diff_property.earliest_diff_value
                if base_previous_diff_value:
                    diff_property.add_value(diff_value=base_previous_diff_value)

    def _apply_relationship_previous_values(
        self, diff_node: DiffNodeIntermediate, base_diff_node: DiffNodeIntermediate
    ) -> None:
        for relationship_key, diff_relationship in diff_node.relationships_by_identifier.items():
            base_diff_relationship = base_diff_node.relationships_by_identifier.get(relationship_key)
            if not base_diff_relationship:
                continue
            for db_id, property_set in diff_relationship.properties_by_db_id.items():
                base_property_set = base_diff_relationship.properties_by_db_id.get(db_id)
                if not base_property_set:
                    continue
                base_diff_property_by_type: dict[DatabaseEdgeType, Optional[DiffRelationshipPropertyIntermediate]] = {
                    DatabaseEdgeType(p.property_type): None for p in property_set
                }
                base_diff_property_by_type[DatabaseEdgeType.IS_RELATED] = None
                latest_diff_property_times_by_type: dict[DatabaseEdgeType, Timestamp] = {}
                for diff_property in property_set:
                    if (
                        diff_property.property_type not in latest_diff_property_times_by_type
                        or diff_property.changed_at > latest_diff_property_times_by_type[diff_property.property_type]
                    ):
                        latest_diff_property_times_by_type[diff_property.property_type] = diff_property.changed_at

                for base_diff_property in base_property_set:
                    prop_type = DatabaseEdgeType(base_diff_property.property_type)
                    if (
                        prop_type not in base_diff_property_by_type
                        or (
                            base_diff_property.status is RelationshipStatus.ACTIVE
                            and base_diff_property.changed_at >= self.from_time
                        )
                        or (
                            prop_type in latest_diff_property_times_by_type
                            and base_diff_property.changed_at > latest_diff_property_times_by_type[prop_type]
                        )
                    ):
                        continue
                    if base_diff_property_by_type[prop_type] is None:
                        base_diff_property_by_type[prop_type] = base_diff_property
                        continue
                    current_property = base_diff_property_by_type.get(prop_type)
                    if current_property and base_diff_property.changed_at < current_property.changed_at:
                        base_diff_property_by_type[prop_type] = base_diff_property
                property_set.update({bdp for bdp in base_diff_property_by_type.values() if bdp})

    def _remove_empty_base_diff_root(self) -> None:
        base_diff_root = self._diff_root_by_branch.get(self.base_branch_name)
        if not base_diff_root:
            return
        for node_diff in base_diff_root.nodes_by_id.values():
            for attribute_diff in node_diff.attributes_by_name.values():
                for property_diff in attribute_diff.properties_by_type.values():
                    ordered_diff_values = property_diff.get_ordered_values_asc()
                    if not ordered_diff_values:
                        continue
                    if ordered_diff_values[-1].changed_at >= self.diff_branched_from_time:
                        return
            for relationship_diff in node_diff.relationships_by_identifier.values():
                for diff_relationship_property_list in relationship_diff.properties_by_db_id.values():
                    for diff_relationship_property in diff_relationship_property_list:
                        if diff_relationship_property.changed_at >= self.diff_branched_from_time:
                            return
        del self._diff_root_by_branch[self.base_branch_name]

    def _finalize(self, include_unchanged: bool) -> None:
        for branch, diff_root_intermediate in self._diff_root_by_branch.items():
            if branch == self.base_branch_name:
                from_time = self.diff_branched_from_time
            else:
                from_time = self.from_time
            self._final_diff_root_by_branch[branch] = diff_root_intermediate.to_diff_root(
                from_time=from_time,
                to_time=self.to_time,
                include_unchanged=include_unchanged,
            )



from infrahub.core.relationship.constraints.peer_kind import RelationshipPeerKindConstraint
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class RelationshipPeerKindConstraintDependency(DependencyBuilder[RelationshipPeerKindConstraint]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipPeerKindConstraint:
        return RelationshipPeerKindConstraint(db=context.db, branch=context.branch)



from typing import TYPE_CHECKING, Any

from graphql import GraphQLResolveInfo
from infrahub_sdk.utils import deep_merge_dict, extract_fields

from infrahub.core.branch.models import Branch
from infrahub.core.constants import BranchSupportType
from infrahub.core.manager import NodeManager
from infrahub.core.schema.relationship_schema import RelationshipSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase

from ..loaders.node import GetManyParams, NodeDataLoader
from ..types import RELATIONS_PROPERTY_MAP, RELATIONS_PROPERTY_MAP_REVERSED

if TYPE_CHECKING:
    from infrahub.core.schema.node_schema import NodeSchema

    from ..initialization import GraphqlContext


class SingleRelationshipResolver:
    def __init__(self) -> None:
        self._data_loader_instances: dict[GetManyParams, NodeDataLoader] = {}

    async def resolve(self, parent: dict, info: GraphQLResolveInfo, **kwargs: Any) -> dict[str, Any]:
        """Resolver for relationships of cardinality=one for Edged responses

        This resolver is used for paginated responses and as such we redefined the requested
        fields by only reusing information below the 'node' key.
        """
        # Extract the InfraHub schema by inspecting the GQL Schema

        node_schema: NodeSchema = info.parent_type.graphene_type._meta.schema  # type: ignore[attr-defined]

        context: GraphqlContext = info.context

        # Extract the name of the fields in the GQL query
        fields = await extract_fields(info.field_nodes[0].selection_set)
        node_fields = fields.get("node", {})
        property_fields = fields.get("properties", {})
        for key, value in property_fields.items():
            mapped_name = RELATIONS_PROPERTY_MAP[key]
            node_fields[mapped_name] = value

        metadata_field_names = {prop_name for prop_name in RELATIONS_PROPERTY_MAP if prop_name != "__typename"}
        requires_relationship_metadata = bool(set(property_fields.keys()) & metadata_field_names)

        # Extract the schema of the node on the other end of the relationship from the GQL Schema
        node_rel = node_schema.get_relationship(info.field_name)

        response: dict[str, Any] = {"node": None, "properties": {}}

        if requires_relationship_metadata:
            node_graph = await self._get_entities_simple(
                db=context.db,
                branch=context.branch,
                at=context.at,
                related_node_ids=context.related_node_ids,
                field_name=info.field_name,
                parent_id=parent["id"],
                source_kind=node_schema.kind,
                rel_schema=node_rel,
                node_fields=node_fields,
                **kwargs,
            )
        else:
            node_graph = await self._get_entities_with_data_loader(
                db=context.db,
                branch=context.branch,
                at=context.at,
                related_node_ids=context.related_node_ids,
                rel_schema=node_rel,
                parent=parent,
                node_fields=node_fields,
            )

        if not node_graph:
            return response
        response["node"] = node_graph

        for key, mapped in RELATIONS_PROPERTY_MAP_REVERSED.items():
            value = node_graph.pop(key, None)
            if value:
                response["properties"][mapped] = value
        return response

    async def _get_entities_simple(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | None,
        related_node_ids: set[str] | None,
        field_name: str,
        parent_id: str,
        source_kind: str,
        rel_schema: RelationshipSchema,
        node_fields: dict[str, Any],
        **kwargs: Any,
    ) -> dict[str, Any] | None:
        filters = {
            f"{field_name}__{key}": value
            for key, value in kwargs.items()
            if "__" in key and value or key in ["id", "ids"]
        }
        async with db.start_session() as dbs:
            objs = await NodeManager.query_peers(
                db=dbs,
                ids=[parent_id],
                source_kind=source_kind,
                schema=rel_schema,
                filters=filters,
                fields=node_fields,
                at=at,
                branch=branch,
                branch_agnostic=rel_schema.branch is BranchSupportType.AGNOSTIC,
                fetch_peers=True,
            )
            if not objs:
                return None
            return await objs[0].to_graphql(db=dbs, fields=node_fields, related_node_ids=related_node_ids)

    async def _get_entities_with_data_loader(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | None,
        related_node_ids: set[str] | None,
        rel_schema: RelationshipSchema,
        parent: dict[str, Any],
        node_fields: dict[str, Any],
    ) -> dict[str, Any] | None:
        try:
            peer_id: str = parent[rel_schema.name][0]["node"]["id"]
        except (KeyError, IndexError):
            return None

        if node_fields and "display_label" in node_fields:
            schema_branch = db.schema.get_schema_branch(name=branch.name)
            display_label_fields = schema_branch.generate_fields_for_display_label(name=rel_schema.peer)
            if display_label_fields:
                node_fields = deep_merge_dict(dicta=node_fields, dictb=display_label_fields)

        if node_fields and "hfid" in node_fields:
            peer_schema = db.schema.get(name=rel_schema.peer, branch=branch, duplicate=False)
            hfid_fields = peer_schema.generate_fields_for_hfid()
            if hfid_fields:
                node_fields = deep_merge_dict(dicta=node_fields, dictb=hfid_fields)

        query_params = GetManyParams(
            fields=node_fields,
            at=at,
            branch=branch,
            include_source=True,
            include_owner=True,
            prefetch_relationships=False,
            account=None,
            branch_agnostic=rel_schema.branch is BranchSupportType.AGNOSTIC,
        )
        if query_params in self._data_loader_instances:
            loader = self._data_loader_instances[query_params]
        else:
            loader = NodeDataLoader(db=db, query_params=query_params)
            self._data_loader_instances[query_params] = loader
        node = await loader.load(key=peer_id)
        if not node:
            return None
        async with db.start_session() as dbs:
            return await node.to_graphql(db=dbs, fields=node_fields, related_node_ids=related_node_ids)



from datetime import timedelta
from typing import Any

import ujson
from infrahub_sdk.protocols import CoreCustomWebhook, CoreStandardWebhook, CoreTransformPython
from prefect import flow
from prefect.automations import AutomationCore
from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName
from prefect.events.actions import RunDeployment
from prefect.events.schemas.automations import EventTrigger, Posture
from prefect.events.schemas.events import ResourceSpecification
from prefect.logging import get_run_logger

from infrahub.core.constants import InfrahubKind, MutationAction
from infrahub.exceptions import NodeNotFoundError
from infrahub.services import services
from infrahub.workflows.catalogue import WEBHOOK_CONFIGURE, WEBHOOK_SEND, WEBHOOK_TRIGGER

from .constants import AUTOMATION_NAME, AUTOMATION_NAME_RUN
from .models import CustomWebhook, SendWebhookData, StandardWebhook, TransformWebhook, Webhook


@flow(name="event-send-webhook", flow_run_name="Send Webhook")
async def send_webhook(model: SendWebhookData) -> None:
    service = services.service
    log = get_run_logger()

    webhook_definition = await service.cache.get(key=f"webhook:active:{model.webhook_id}")
    if not webhook_definition:
        log.warning("Webhook not found")
        raise NodeNotFoundError(
            node_type="Webhook", identifier=model.webhook_id, message="The requested Webhook was not found"
        )

    webhook_data = ujson.loads(webhook_definition)
    payload: dict[str, Any] = {"event_type": model.event_type, "data": model.event_data, "service": service}
    webhook_map: dict[str, type[Webhook]] = {
        "standard": StandardWebhook,
        "custom": CustomWebhook,
        "transform": TransformWebhook,
    }
    webhook_class = webhook_map[webhook_data["webhook_type"]]
    payload.update(webhook_data["webhook_configuration"])
    webhook = webhook_class(**payload)
    await webhook.send()

    log.info("Successfully sent webhook")


@flow(name="webhook-trigger-actions", flow_run_name="Trigger configured webhooks")
async def trigger_webhooks(event_type: str, event_data: str) -> None:
    service = services.service
    payload: dict = ujson.loads(event_data)

    webhooks = await service.cache.list_keys(filter_pattern="webhook:active:*")
    for webhook in webhooks:
        webhook_id = webhook.split(":")[-1]
        model = SendWebhookData(webhook_id=webhook_id, event_type=event_type, event_data=payload)
        await service.workflow.submit_workflow(workflow=WEBHOOK_SEND, parameters={"model": model})


@flow(name="webhook-setup-automations", flow_run_name="Configuration webhook automation and populate cache")
async def configure_webhooks() -> None:
    service = services.service
    log = get_run_logger()

    log.debug("Refreshing webhook configuration")
    standard_webhooks = await service.client.all(kind=CoreStandardWebhook)
    custom_webhooks = await service.client.all(kind=CoreCustomWebhook)

    expected_webhooks = []
    for webhook in standard_webhooks:
        webhook_key = f"webhook:active:{webhook.id}"
        expected_webhooks.append(webhook_key)
        standard_payload = {
            "webhook_type": "standard",
            "webhook_configuration": {
                "url": webhook.url.value,
                "shared_key": webhook.shared_key.value,
                "validate_certificates": webhook.validate_certificates.value,
            },
        }
        await service.cache.set(key=webhook_key, value=ujson.dumps(standard_payload))

    for webhook in custom_webhooks:
        webhook_key = f"webhook:active:{webhook.id}"
        expected_webhooks.append(webhook_key)
        payload: dict[str, Any] = {
            "webhook_type": "custom",
            "webhook_configuration": {
                "url": webhook.url.value,
                "validate_certificates": webhook.validate_certificates.value,
            },
        }
        if webhook.transformation.id:
            transform = await service.client.get(
                kind=CoreTransformPython,
                id=webhook.transformation.id,
                prefetch_relationships=True,
                populate_store=True,
                include=["name", "class_name", "file_path", "repository"],
            )
            payload["webhook_type"] = "transform"
            payload["webhook_configuration"]["transform_name"] = transform.name.value
            payload["webhook_configuration"]["transform_class"] = transform.class_name.value
            payload["webhook_configuration"]["transform_file"] = transform.file_path.value
            payload["webhook_configuration"]["repository_id"] = transform.repository.id
            payload["webhook_configuration"]["repository_name"] = transform.repository.peer.name.value

        await service.cache.set(key=webhook_key, value=ujson.dumps(payload))

    cached_webhooks = await service.cache.list_keys(filter_pattern="webhook:active:*")
    for cached_webhook in cached_webhooks:
        if cached_webhook not in expected_webhooks:
            await service.cache.delete(key=cached_webhook)

    has_webhooks = bool(expected_webhooks)

    async with get_client(sync_client=False) as client:
        deployments = {
            item.name: item
            for item in await client.read_deployments(
                deployment_filter=DeploymentFilter(
                    name=DeploymentFilterName(
                        any_=[
                            WEBHOOK_TRIGGER.name,
                        ]
                    )
                )
            )
        }
        if WEBHOOK_TRIGGER.name not in deployments:
            raise ValueError("Unable to find the deployment for WEBHOOK_TRIGGER")

        deployment_id_webhook_trigger = deployments[WEBHOOK_TRIGGER.name].id

        webhook_configure_automation = await client.find_automation(id_or_name=AUTOMATION_NAME_RUN)

        if not has_webhooks:
            if webhook_configure_automation:
                await client.delete_automation(automation_id=webhook_configure_automation.id)
            return

        automation = AutomationCore(
            name=AUTOMATION_NAME_RUN,
            description="Trigger all configured webhooks on mutations",
            enabled=True,
            trigger=EventTrigger(
                posture=Posture.Reactive,
                expect={"infrahub.node.*"},
                within=timedelta(0),
                match=ResourceSpecification(
                    {
                        "infrahub.node.action": MutationAction.available_types(),
                    }
                ),
                threshold=1,
            ),
            actions=[
                RunDeployment(
                    source="selected",
                    deployment_id=deployment_id_webhook_trigger,
                    parameters={
                        "event_type": "{{ event.resource['infrahub.node.kind'] }}.{{ event.resource['infrahub.node.action'] }}",
                        "event_data": "{{ event.payload['data'] | tojson }}",
                    },
                    job_variables={},
                ),
            ],
        )

        if webhook_configure_automation:
            await client.update_automation(automation_id=webhook_configure_automation.id, automation=automation)
            log.info(f"{AUTOMATION_NAME_RUN} Updated")
        else:
            await client.create_automation(automation=automation)
            log.info(f"{AUTOMATION_NAME_RUN} Created")


@flow(name="webhook-setup-configuration-trigger", flow_run_name="Setup automations for webhooks")
async def trigger_webhook_configuration() -> None:
    log = get_run_logger()

    async with get_client(sync_client=False) as client:
        deployments = {
            item.name: item
            for item in await client.read_deployments(
                deployment_filter=DeploymentFilter(
                    name=DeploymentFilterName(
                        any_=[
                            WEBHOOK_CONFIGURE.name,
                        ]
                    )
                )
            )
        }
        if WEBHOOK_CONFIGURE.name not in deployments:
            raise ValueError("Unable to find the deployment for WEBHOOK_CONFIGURE")

        deployment_id_webhook_setup = deployments[WEBHOOK_CONFIGURE.name].id

        webhook_configure_automation = await client.find_automation(id_or_name=AUTOMATION_NAME)

        automation = AutomationCore(
            name=AUTOMATION_NAME,
            description="Trigger actions on schema update event",
            enabled=True,
            trigger=EventTrigger(
                posture=Posture.Reactive,
                expect={"infrahub.node.*"},
                within=timedelta(0),
                match=ResourceSpecification(
                    {
                        "infrahub.node.kind": [InfrahubKind.WEBHOOK, InfrahubKind.STANDARDWEBHOOK],
                    }
                ),
                threshold=1,
            ),
            actions=[
                RunDeployment(
                    source="selected",
                    deployment_id=deployment_id_webhook_setup,
                    parameters={},
                    job_variables={},
                ),
            ],
        )

        if webhook_configure_automation:
            await client.update_automation(automation_id=webhook_configure_automation.id, automation=automation)
            log.info(f"{AUTOMATION_NAME} Updated")
        else:
            await client.create_automation(automation=automation)
            log.info(f"{AUTOMATION_NAME} Created")



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterator, Optional, TypeVar, Union

import ujson
from neo4j.graph import Node as Neo4jNode
from neo4j.graph import Path as Neo4jPath
from neo4j.graph import Relationship as Neo4jRelationship
from opentelemetry import trace

from infrahub import config
from infrahub.core.constants import PermissionLevel
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import QueryError

if TYPE_CHECKING:
    from neo4j import Record
    from typing_extensions import Self

    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase


RETURN_TYPE = TypeVar("RETURN_TYPE")


def sort_results_by_time(results: list[QueryResult], rel_label: str) -> list[QueryResult]:
    """Sort a list of QueryResult based on the to and from fields on given relationship.

    To sort the results, we are generating an ID per item
        The ID is derived from the timestamp of the last action and we multiply it by 1000
        To differentiate two records that have been updated at the same time
        We are adding more weight (500) to the record for which the last action was to set "from"
         versus a record with "from" and "to" set.
    """

    results_dict = {}

    for result in results:
        from_time = result.get(rel_label).get("from")
        to_time = result.get(rel_label).get("to")

        if not to_time:
            record_id = Timestamp(from_time).to_timestamp() * 1000 + 500
        else:
            record_id = Timestamp(to_time).to_timestamp() * 1000

        results_dict[record_id] = result

    return [value for _, value in sorted(results_dict.items(), reverse=False)]


class QueryElementType(Enum):
    NODE = "node"
    RELATIONSHIP = "relationship"


class QueryRelDirection(Enum):
    BIDIR = "bidirectional"
    INBOUND = "inbound"
    OUTBOUND = "outbound"


@dataclass
class QueryElement:
    type: QueryElementType
    name: Optional[str] = None
    labels: Optional[list[str]] = None
    params: Optional[dict] = None

    def __str__(self) -> str:
        main_str = "%s%s%s" % (self.name or "", self.labels_as_str, self.params_as_str)
        if self.type == QueryElementType.NODE:
            return "(%s)" % main_str
        return "[%s]" % main_str

    @property
    def labels_as_str(self) -> str:
        if not self.labels:
            return ""

        return ":" + ":".join(self.labels)

    @property
    def params_as_str(self) -> str:
        if not self.params:
            return ""

        params_list = []
        for key, value in self.params.items():
            if isinstance(value, str) and not value.startswith("$"):
                value_str = f'"{value}"'
            else:
                value_str = value

            params_list.append(f"{key}: {value_str}")

        return " { " + ",".join(params_list) + " }"


@dataclass
class QueryNode(QueryElement):
    type: QueryElementType = QueryElementType.NODE


@dataclass
class QueryRel(QueryElement):
    type: QueryElementType = QueryElementType.RELATIONSHIP
    direction: QueryRelDirection = QueryRelDirection.BIDIR
    length_min: int = 1
    length_max: Optional[int] = None

    def __str__(self) -> str:
        length_str = ""
        if self.length_max:
            length_str = "*%s..%s" % (
                self.length_min,
                self.length_max,
            )

        main_str = "[%s%s%s%s]" % (
            self.name or "",
            self.labels_as_str,
            length_str,
            self.params_as_str,
        )

        if self.direction == QueryRelDirection.INBOUND:
            return "<-%s-" % main_str
        if self.direction == QueryRelDirection.OUTBOUND:
            return "-%s->" % main_str

        return "-%s-" % main_str


class QueryType(Enum):
    READ = "read"
    WRITE = "write"


def cleanup_return_labels(labels: list[str]) -> list[str]:
    """Cleanup a list of return labels by checking if there is an alias defined.
    if an alias is defined with `value AS alias` we extract just the alias from the label
    """
    clean_labels = []
    for label in labels:
        idx = label.lower().find(" as ")
        if idx > 0:
            clean_idx = idx + 4
            clean_label = label[clean_idx:]
            clean_labels.append(clean_label.strip())
        else:
            clean_labels.append(label)

    return clean_labels


class QueryResult:
    def __init__(self, data: list[Union[Neo4jNode, Neo4jRelationship, list[Neo4jNode]]], labels: list[str]):
        self.data = data
        self.labels = labels
        self.branch_score: int = 0
        self.time_score: int = 0
        self.permission_score = PermissionLevel.DEFAULT
        self.has_deleted_rels: bool = False

        self.calculate_branch_score()
        self.calculate_time_score()
        self.check_rels_status()

    def calculate_branch_score(self) -> None:
        """The branch score is a simple way to order and classify multiple responses for the same branch.
        If the branch name is not the default branch it will get a higher score
        """
        self.branch_score = 0

        for rel in self.get_rels():
            branch_level = rel.get("branch_level", None)

            if not branch_level:
                continue

            self.branch_score += branch_level

    def calculate_time_score(self) -> None:
        """The time score look into the to and from time all relationships
        if the 'to' field is not defined
        """
        self.time_score = 0

        for rel in self.get_rels():
            branch_name = rel.get("branch", None)

            if not branch_name:
                continue

            to_time = rel.get("to", None)

            if to_time:
                self.time_score += 1
            else:
                self.time_score += 2

    def check_rels_status(self) -> None:
        """Check if some relationships have the status deleted and update the flag `has_deleted_rels`"""
        for rel in self.get_rels():
            if rel.get("status", None) == "deleted":
                self.has_deleted_rels = True
                return

    def _get(self, label: str) -> Union[Neo4jNode, Neo4jRelationship, list[Neo4jNode]]:
        if label not in self.labels:
            raise ValueError(f"{label} is not a valid value for this query, must be one of {self.labels}")

        return_id = self.labels.index(label)
        return self.data[return_id]

    def get(self, label: str) -> Union[Neo4jNode, Neo4jRelationship]:
        return self._get(label=label)

    def get_as_str(self, label: str) -> Optional[str]:
        item = self._get(label=label)
        if item:
            return str(item)
        return None

    def get_as_optional_type(self, label: str, return_type: Callable[..., RETURN_TYPE]) -> Optional[RETURN_TYPE]:
        """Return a label as a given type.

        For example if an integer is needed the caller would use:
        .get_as_optional_type(label="name_of_label", return_type=int)
        """
        item = self._get(label=label)
        if item is not None:
            return return_type(item)
        return None

    def get_as_type(self, label: str, return_type: Callable[..., RETURN_TYPE]) -> RETURN_TYPE:
        """Return a label as a given type.

        For example if an integer is needed the caller would use:
        .get_as_type(label="name_of_label", return_type=int)
        """
        item = self._get(label=label)

        return return_type(item)

    def get_node_collection(self, label: str) -> list[Neo4jNode]:
        entry = self._get(label=label)
        if isinstance(entry, list):
            return entry
        raise ValueError(f"{label} is not a collection use .get_node() or .get()")

    def get_nested_node_collection(self, label: str) -> list[list[Neo4jNode]]:
        entry = self._get(label=label)
        if isinstance(entry, list):
            return entry
        raise ValueError(f"{label} is not a collection use .get_node() or .get()")

    def get_node(self, label: str) -> Neo4jNode:
        node = self.get(label=label)
        if isinstance(node, Neo4jNode):
            return node
        raise ValueError(f"{label} is not a Node")

    def get_rel(self, label: str) -> Neo4jRelationship:
        rel = self.get(label=label)
        if isinstance(rel, Neo4jRelationship):
            return rel
        raise ValueError(f"{label} is not a Relationship")

    def get_rels(self) -> Generator[Neo4jRelationship, None, None]:
        """Return all relationships."""

        for item in self.data:
            if isinstance(item, Neo4jRelationship):
                yield item

    def get_nodes(self) -> Generator[Neo4jNode, None, None]:
        """Return all nodes."""
        for item in self.data:
            if isinstance(item, Neo4jNode):
                yield item

    def get_path(self, label: str) -> Neo4jPath:
        path = self.get(label=label)
        if isinstance(path, Neo4jPath):
            return path
        raise ValueError(f"{label} is not a Path")

    def get_paths(self, label: str) -> Generator[Neo4jPath, None, None]:
        for path in self.get(label=label):
            if isinstance(path, Neo4jPath):
                yield path

    def __iter__(self) -> Iterator:
        yield from self.data


@dataclass
class QueryStats:
    stats: list[QueryStat] = field(default_factory=list)

    def add(self, data: Optional[dict[str, Any]]) -> None:
        if data:
            self.stats.append(QueryStat.from_metadata(data))

    def get_counter(self, name: str) -> int:
        cnt = 0
        for stat in self.stats:
            if not hasattr(stat, name):
                raise ValueError(f"Counter {name} is not available")
            cnt += getattr(stat, name)

        return cnt


@dataclass
class QueryStat:
    contains_updates: bool = False
    labels_added: Optional[int] = None
    labels_removed: Optional[int] = None
    nodes_created: Optional[int] = None
    nodes_deleted: Optional[int] = None
    properties_set: Optional[int] = None
    relationships_created: Optional[int] = None
    relationships_deleted: Optional[int] = None

    @classmethod
    def from_metadata(cls, data: dict[str, Any]) -> Self:
        data = {key.replace("-", "_"): value for key, value in data.items()}
        return cls(**data)


class Query(ABC):
    name: str = "base-query"
    type: QueryType

    raise_error_if_empty: bool = False
    insert_return: bool = True
    insert_limit: bool = True

    def __init__(
        self,
        branch: Optional[Branch] = None,
        at: Optional[Union[Timestamp, str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[list[str]] = None,
        branch_agnostic: bool = False,
    ):
        if branch:
            self.branch = branch

        self.branch_agnostic = branch_agnostic

        if not hasattr(self, "at"):
            self.at = Timestamp(at)

        if not self.at:
            self.at = Timestamp(at)

        self.limit = limit
        self.offset = offset
        self.order_by = order_by

        # Initialize internal variables
        self.params: dict = {}
        self.query_lines: list[str] = []
        self.return_labels: list[str] = []
        self.results: list[QueryResult] = []

        self.has_been_executed: bool = False
        self.has_errors: bool = False

        self.stats: QueryStats = QueryStats()

    def update_return_labels(self, value: Union[str, list[str]]) -> None:
        if isinstance(value, str) and value not in self.return_labels:
            self.return_labels.append(value)
            return
        if isinstance(value, list):
            for item in value:
                self.update_return_labels(value=item)

    @classmethod
    async def init(
        cls,
        db: InfrahubDatabase,
        branch: Optional[Branch] = None,
        at: Optional[Union[Timestamp, str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **kwargs: Any,
    ) -> Self:
        query = cls(branch=branch, at=at, limit=limit, offset=offset, **kwargs)

        await query.query_init(db=db, **kwargs)

        return query

    @abstractmethod
    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        raise NotImplementedError

    def get_context(self) -> dict[str, str]:
        """Provide additional context for this query, beyond the name.
        Right now it's mainly used to add more labels to the metrics."""
        return {}

    def add_to_query(self, query: Union[str, list[str]]) -> None:
        """Add a new section at the end of the query.

        A string with multiple lines will be broken down into multiple entries in self.query_lines
        Trailing and leading spaces per line will be removed."""

        if isinstance(query, list):
            for item in query:
                self.add_to_query(query=item)
        else:
            self.query_lines.extend([line.strip() for line in query.split("\n") if line.strip()])

    def add_subquery(self, subquery: str, with_clause: Optional[str] = None) -> None:
        self.add_to_query("CALL {")
        self.add_to_query(subquery)
        self.add_to_query("}")
        if with_clause:
            self.add_to_query(f"WITH {with_clause}")

    def get_query(
        self,
        var: bool = False,
        inline: bool = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> str:
        # Make a local copy of the _query_lines
        limit = limit or self.limit
        offset = offset or self.offset
        tmp_query_lines = self.query_lines.copy()

        if self.insert_return:
            tmp_query_lines.append("RETURN " + ",".join(self.return_labels))

        if self.order_by:
            tmp_query_lines.append("ORDER BY " + ",".join(self.order_by))

        if offset and self.insert_limit:
            tmp_query_lines.append(f"SKIP {offset}")

        if limit and self.insert_limit:
            tmp_query_lines.append(f"LIMIT {limit}")

        query_str = "\n".join(tmp_query_lines)

        if var and not inline:
            return "\n" + self.get_params_for_shell() + "\n\n" + query_str
        if var and inline:
            return self.insert_variables_in_query(query=query_str, variables=self.params)

        return query_str

    def get_count_query(self, var: bool = False) -> str:
        tmp_query_lines = self.query_lines.copy()
        tmp_query_lines.append("RETURN count(*) as count")
        query_str = "\n".join(tmp_query_lines)

        if not var:
            return query_str

        return self.insert_variables_in_query(query=query_str, variables=self.params)

    @classmethod
    def insert_variables_in_query(cls, query: str, variables: dict) -> str:
        """Search for all the variables in a Query string and replace each variable with its value."""

        def prep_value(v: Any) -> str:
            if isinstance(v, (int, list)):
                return str(v)
            return f'"{v}"'

        # Sort the list of variables first to ensure the longest will be processed first
        vars_list = list(variables.items())
        vars_list.sort(key=lambda x: len(x[0]), reverse=True)
        for key, value in vars_list:
            if isinstance(value, dict):
                # First try to insert individual element of the dict as var
                sub_vars = {f"{key}.{sub_key}": sub_value for sub_key, sub_value in value.items()}
                query = cls.insert_variables_in_query(query=query, variables=sub_vars)

                # Then replace the entire object if nothing else was found
                value_items = [f"{key1}: {prep_value(value1)}" for key1, value1 in value.items()]
                value_str = "{ " + ", ".join(value_items) + " }"
                query = query.replace(f"${key}", value_str)
            else:
                query = query.replace(f"${key}", prep_value(value))

        return query

    def get_params_for_shell(self) -> str:
        if config.SETTINGS.database.db_type.value == "memgraph":
            return ujson.dumps(self.params)

        return self._get_params_for_neo4j_shell()

    def _get_params_for_neo4j_shell(self) -> str:
        """Generate string to define some parameters in Neo4j browser interface.
        It's especially useful to later execute a query that includes some variables.

        The params string must be executed on its own window in Neo4j, before executing the query.
        """

        params = []

        for key, value in self.params.items():
            if isinstance(value, (int, list)):
                params.append(f"{key}: {str(value)}")
            else:
                params.append(f'{key}: "{value}"')

        return ":params { " + ", ".join(params) + " }"

    @trace.get_tracer(__name__).start_as_current_span("Query.execute")
    async def execute(self, db: InfrahubDatabase) -> Self:
        # Ensure all mandatory params have been provided
        # Ensure at least 1 return obj has been defined

        if config.SETTINGS.miscellaneous.print_query_details:
            self.print(include_var=True)

        query_str = self.get_query()

        if self.type == QueryType.READ:
            if self.limit or self.offset:
                results = await db.execute_query(
                    query=query_str, params=self.params, name=self.name, context=self.get_context(), type=self.type
                )
            else:
                results = await self.query_with_size_limit(db=db)

        elif self.type == QueryType.WRITE:
            results, metadata = await db.execute_query_with_metadata(
                query=query_str, params=self.params, name=self.name, context=self.get_context(), type=self.type
            )
            if "stats" in metadata:
                self.stats.add(metadata.get("stats"))
        else:
            raise ValueError(f"unknown value for {self.type}")

        if not results and self.raise_error_if_empty:
            raise QueryError(query=query_str, params=self.params)

        clean_labels = cleanup_return_labels(self.return_labels)
        self.results = [QueryResult(data=result, labels=clean_labels) for result in results]
        self.has_been_executed = True

        return self

    async def query_with_size_limit(self, db: InfrahubDatabase) -> list[Record]:
        query_limit = config.SETTINGS.database.query_size_limit
        offset = 0
        results: list[Record] = []
        remaining = True
        while remaining:
            offset_results, metadata = await db.execute_query_with_metadata(
                query=self.get_query(limit=query_limit, offset=offset),
                params=self.params,
                name=self.name,
                context=self.get_context(),
                type=self.type,
            )
            if "stats" in metadata:
                self.stats.add(metadata.get("stats"))
            results.extend(offset_results)
            offset += query_limit

            if len(offset_results) < query_limit:
                remaining = False

        return results

    async def count(self, db: InfrahubDatabase) -> int:
        """Count the number of results matching a READ query.
        OFFSET and LIMIT are automatically excluded when counting.
        """

        if self.type == QueryType.WRITE:
            raise TypeError("Unable to count the number of response on a Write query.")
        if self.type != QueryType.READ:
            raise ValueError(f"unknown value for {self.type}")

        results = await db.execute_query(
            query=self.get_count_query(), params=self.params, name=f"{self.name}_count", type=QueryType.READ
        )

        if not results and self.raise_error_if_empty:
            raise QueryError(query=self.get_count_query(), params=self.params)

        return results[0][0]

    def get_result(self) -> Optional[QueryResult]:
        """Return a single Result."""

        if not self.has_been_executed:
            return None

        if self.num_of_results == 1:
            return self.results[0]

        try:
            return next(self.get_results())
        except StopIteration:
            return None

    def get_results(self) -> Generator[QueryResult, None, None]:
        """Get all the results sorted by score."""

        score_idx = {}
        for idx, result in enumerate(self.results):
            score_idx[idx] = result.branch_score

        for idx, _ in sorted(score_idx.items(), key=lambda x: x[1], reverse=True):
            yield self.results[idx]

    def get_results_group_by(self, *args: Any) -> Generator[QueryResult, None, None]:
        """Return results group by the labels and attributes provided and filtered by scored.

        Examples:
            get_results_group_by(("n", "uuid"), ("a", "name")):
        """

        attrs_info = defaultdict(list)

        # Extract all attrname and relationships on all branches
        for idx, result in enumerate(self.results):
            identifier = []
            for label, attribute in args:
                node = result.get(label)
                if hasattr(node, attribute):
                    identifier.append(getattr(node, attribute))
                else:
                    identifier.append(node.get(attribute, None))

            info = {
                "idx": idx,
                "branch_score": result.branch_score,
                "time_score": result.time_score,
                "deleted": result.has_deleted_rels,
            }
            attrs_info[tuple(identifier)].append(info)

        for values in attrs_info.values():
            attr_info = sorted(
                values, key=lambda i: (i["branch_score"], i["time_score"], not i["deleted"]), reverse=True
            )[0]
            if attr_info["deleted"]:
                continue

            yield self.results[attr_info["idx"]]

    @property
    def num_of_results(self) -> int:
        if not self.has_been_executed:
            raise ValueError("The query hasn't been executed yet")

        return len([result for result in self.results if not result.has_deleted_rels])

    def print_table(self) -> None:
        # pylint: disable=import-outside-toplevel

        from rich.console import Console
        from rich.table import Table

        console = Console()

        table = Table(title=f"Query {self.name} : params: {self.params}")

        for label in self.return_labels:
            # table.add_column("Name", justify="right", style="cyan", no_wrap=True)
            table.add_column(label)

        for result in self.results:
            table.add_row(*[str(result.get(label)) for label in self.return_labels])

        console.print(table)

    def print(self, include_var: bool = False) -> None:
        # pylint: disable=import-outside-toplevel
        from rich import print as rprint

        print("-------------------------------------------------------")
        print(self.get_query(var=include_var))
        if self.params:
            rprint(self.params)
